"""
File Purpose: making a new eppic input deck file, possibly based on existing file.
"""
import ast
import os

import numpy as np
import pandas as pd  # for pretty table of timescales

from .eppic_io_tools import (
    update_eppic_i_file, get_updated_eppic_i_str,
    _get_update_eppic_i_values_flat,
)
from ...errors import DimensionError, InputError
from ...tools import (
    format_docstring,
    alias,
    InDir, using_attrs,
    UNSET,
)
from ...defaults import DEFAULTS


''' --------------------- EppicParam --------------------- '''

class EppicParam():
    '''a value for a single eppic parameters.
    Also with instructions for converting to str, and info about the old value.

    value: any
        value to use for this parameter.
    old: any
        old value for this parameter.
    fmt: str
        instructions for formatting the value.

    when this object is converted to str, will return fmt.format(value).
    '''
    def __init__(self, value, *, old=None, fmt='{:}'):
        self.value = value
        self.old = old
        self.fmt = fmt

    def fmt_str(self, value):
        '''returns self.fmt.format(value), or '0' if value==0.'''
        return '0' if value==0 else self.fmt.format(value)

    def str_value(self):
        '''returns self.fmt_str(self.value)'''
        return self.fmt_str(self.value)

    def str_old(self):
        '''returns self.fmt_str(self.old), or 'None' if self.old is None'''
        return 'None' if self.old is None else self.fmt_str(self.old)

    def eval_str_value(self):
        '''returns result of evaluating self.str_value().
        Might be different from value, due to rounding when converting to string.
        '''
        return ast.literal_eval(self.str_value())

    def eval_str_old(self):
        '''returns result of evaluating self.str_old().
        Might be different from old, due to rounding when converting to string.
        '''
        return ast.literal_eval(self.str_old())

    # # # DISPLAY # # #
    def __str__(self):
        return self.str_value()

    def __repr__(self):
        str_value = self.str_value()
        contents = [str_value]
        str_old = self.str_old()
        if str_old != str_value:
            contents.append(f'old={self.str_old()}')
        return f'{type(self).__name__}({", ".join(contents)})'


''' --------------------- EppicChooseParams --------------------- '''

class EppicChooseParamsBase():
    '''class to help with choosing parameters for eppic.i file.

    Probably, use EppicChooseParams instead.
    
    values: dict
        {varname: value} for all pairs to update.
        Can also provide {dist number: dict of {varname: value} pairs to update for this dist}
            (inside of dist number dicts, varname can end with dist number or not;
            assume it does end with dist number in the file itself though)
            In this case, will be immediately converted to values_flat format.
    values_as_kw: optionally, provide additional values as kwargs.
    '''
    # # # CREATION # # #
    def __init__(self, values=dict(), *, dirname=None, **values_as_kw):
        self.values = self.values_flat(values, **values_as_kw)
        for k, v in self.values.items():
            if not isinstance(value, EppicParam):
                self.values[k] = EppicParam(v)
        self.dirname = os.path.abspath(os.curdir if dirname is None else dirname)

    # # # INSPECTION # # #
    @staticmethod
    def values_flat(values=dict(), **values_as_kw):
        '''return these values as a 'flat' dict, suitable for using to update eppic.i file.
        This is a dict of {key: value} pairs, appending dist number to key name where appropriate.
        '''
        return _get_update_eppic_i_values_flat({**values, **values_as_kw})

    # # # ADJUST VALUES # # #
    def update(self, values=dict(), **values_as_kw):
        '''update self.values with values. returns self'''
        values_flat = self.values_flat(values, **values_as_kw)
        self.values.update(values_flat)
        return self

    def setdefault(self, values=dict(), **values_as_kw):
        '''setdefault for self.values. returns self.
        Like self.update, but doesn't overwrite existing values.
        '''
        values_flat = self.values_flat(values, **values_as_kw)
        for k, v in values_flat.items():
            self.values.setdefault(k, v)
        return self

    # # # DICT-LIKE # # #
    def __getitem__(self, key):
        return self.values[key]
    def __setitem__(self, key, value):
        self.values[key] = value
    def __delitem__(self, key):
        del self.values[key]
    def __iter__(self):
        return iter(self.values)
    def __len__(self):
        return len(self.values)
    def __contains__(self, key):
        return key in self.values
    def items(self):
        return self.values.items()
    def keys(self):
        return self.values.keys()
    # don't def values()! we already have self.values...

    # # # WRITING TO FILE # # #
    def write(self, src='eppic.i', dst='eppic_updated.i', *,
              exists_ok=False, missing_values_ok=True, comment='previously={old}',
              **kw):
        '''write updated eppic.i file to dst.
        src: str
            path to source file to use as template.
        dst: str
            path to destination file to write.
        exists_ok: bool, default False
            whether it is okay for dst to already exist.
        missing_values_ok: bool
            whether it is okay for some of the varnames in values to not exist in eppic.i
        comment: None or str
            if not None, append comment as a comment to all updated lines.
            Also, this comment will be hit by .format(old=old value)

        additional kw passed to update_eppic_i_file.

        return abspath to dst.
        '''
        kw.update(exists_ok=exists_ok, missing_values_ok=missing_values_ok, comment=comment)
        with InDir(self.dirname):
            return update_eppic_i_file(src, dst, self.values, **kw)

    def str_from(self, src='eppic.i', missing_values_ok=True, comment='previously={old}', **kw):
        '''return str to use for the entire contents of the new, updated eppic.i file.
        src: str
            path to source file to use as template.
        missing_values_ok: bool
            whether it is okay for some of the varnames in values to not exist in eppic.i
        comment: None or str
            if not None, append comment as a comment to all updated lines.
            Also, this comment will be hit by .format(old=old value)

        additional kw passed to get_updated_eppic_i_str
        '''
        return get_updated_eppic_i_str(src, None, self.values, **kw)

    def __repr__(self):
        return f'{type(self).__name__}(values={self.values}, dirname={self.dirname!r})'


class EppicChooseParams(EppicChooseParamsBase):
    '''class to help with choosing parameters for eppic.i file.
    
    ec: EppicCalculator object
        the EppicCalculator to use for getting values and directory of original eppic.i file.
    var: None, str, or list of strings
        tells which getters to use when getting values. See self.GETTERS.keys() for options.
        None --> use all vars in self.GETTERS.
        str --> use this var.
        list --> use these vars.
        Use [] to use no vars, i.e. initialize only with user-provided values.
        if getting vars gives keys also provided in ``values``, use the value from ``values`` instead.
    values: dict
        {varname: value} for pairs to update in new eppic.i file.
        Can also provide {dist number: dict of {varname: value} pairs to update for this dist}
            (inside of dist number dicts, varname can end with dist number or not;
            assume it does end with dist number in the file itself though)
            Will be immediately converted to self.values_flat format.
    additional kwargs:
        passed as self(var, **kw_call) when determining initial values.

    Examples:
        Basic usage:
            choices = EppicChooseParams(eppic_calculator_object)  # create instance;
            # ^ will also fill choices with defaults based on calculations,
            #   e.g. including vtherm calculated from temperatures, and dx from ldebye from electrons.
            choices.update(...)   # any choices of params you specifically want to set can go here.
            choices.write(...)  # write a new eppic.i file based on the original, and choices.

        Other common use case - learn info about timescales:
            choices.timescales()  # EppicTimescales object
    '''
    def __init__(self, ec, var=None, *, values=dict(), **kw_call_self):
        '''create EppicChooseParameters by reading values from EppicCalculator.get_vals_for_inputs()'''
        self.ec = ec
        dirname = getattr(self.ec.input_deck, 'dirname', None)
        super().__init__(values, dirname=dirname)
        get_var_values = self(var=var, **kw_call_self)
        self.setdefault(get_var_values)

    # # # CALCULATING & CHECKING VALUES # # #
    GETTERS = dict()  # dict of {param name: method name} to use for getting values.
    CHECKERS = dict()  # dict of {check name: method name} to use for checking values.

    checks_mode = 'warn'   # default mode for doing checks. See help(self.__call__) for details.

    get = alias('__call__')

    def __call__(self, var=None, *, precision=UNSET, checks=UNSET, units='raw', **kw):
        '''return dict of {key: value} to use in input deck, based on current values in self.ec.
        (Ignores current values of self. Consider using: self.setdefault(self()).)
        Or, checks current values in self.ec for "reasonable-ness".

        var: None, str, or list of strings
            tells which getters & checkers to include.
            See self.GETTERS.keys() and self.CHECKERS.keys() for options.
            None --> use all vars in self.GETTERS and all vars in self.CHECKERS.
            str --> use this var.
            list --> use these vars.
        precision: UNSET, dict, or int
            digits of precision for exponential values when converting to strings.
            UNSET --> use each var's default, defined within its getter function.
            dict --> {{var: precision to use for var}}, for var in self.GETTERS.
        checks: UNSET, bool, 'warn', or 'crash'
            whether to check values for "reasonable-ness", and how to behave if a check fails.
            UNSET --> use self.checks_mode.
            False --> don't check.
            True --> equivalent to 'warn'
            'warn' --> print warning message for any failed checks.
            'crash' --> raise AssertionError for any failed checks.
        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        # # bookkeeping # #
        # var
        if var is None:
            var = list(self.GETTERS.keys())
            var.extend(v for v in self.CHECKERS.keys() if not v in var)
        else:
            if isinstance(var, str):
                var = [var]
            for v in var:
                if not (v in self.GETTERS or v in self.CHECKERS):
                    errmsg = (f'unknown var {v!r}! Expected one of self.GETTERS.keys(): {list(self.GETTERS.keys())} '
                              f'or self.CHECKERS.keys(): {list(self.CHECKERS.keys())}')
                    raise KeyError(errmsg)
        # precision
        if precision is UNSET:
            precision = dict()
        elif isinstance(precision, int):
            precision = {v: precision for v in var}
        # looping
        values = dict()
        with self.ec.using(units=units, **kw):
            for v in var:
                if v in self.GETTERS:
                    getter = getattr(self, self.GETTERS[v])
                    kw_v = dict(units=units, **kw)
                    if v in precision:
                        kw_v['precision'] = precision[v]
                    values.update(getter(**kw_v))
                if v in self.CHECKERS:
                    self.check(v, mode=checks)
        return values

    def check(self, var=None, *, mode=UNSET, **kw_checker):
        '''check current values in self.ec for "reasonable-ness".

        var: None, str, or list of strings
            tells which checkers to include. See self.CHECKERS.keys() for options.
            None --> use all vars in self.CHECKERS.
            str --> use this var.
            list --> use these vars.
        mode: UNSET, bool, 'warn', or 'crash'
            whether to check values for "reasonable-ness", and how to behave if a check fails.
            UNSET --> use self.checks_mode.
            False --> don't check.
            True --> equivalent to 'warn'
            'warn' --> print warning message for any failed checks.
            'crash' --> raise AssertionError for any failed checks.

        additional kwargs are passed to every used checker.

        returns dict {v: result from check_v} for each v in var.
        '''
        if var is None:
            var = list(self.CHECKERS.keys())
        else:
            if isinstance(var, str):
                var = [var]
            for v in var:
                if not (v in self.CHECKERS):
                    errmsg = (f'unknown var {v!r}! Expected one of self.CHECKERS.keys(): {list(self.CHECKERS.keys())}')
                    raise KeyError(errmsg)
        # checks_mode
        if mode is UNSET:
            checks_mode = self.checks_mode
        elif mode in (False, True, 'warn', 'crash'):
            checks_mode = mode
        else:
            raise InputError(f'invalid mode={mode!r}. Must be UNSET, False, True, "warn", or "crash".')
        if not mode:
            return dict()  # nothing to check.
        # looping
        result = dict()
        with using_attrs(self, checks_mode=checks_mode):
            for v in var:
                checker = getattr(self, self.CHECKERS[v])
                result[v] = checker(**kw_checker)
        return result

    # # GETTERS # #
    def _param_getter(f, getters=GETTERS):
        '''return f after putting f.__name__ into getters
        assert f.__name__ looks like 'get_{var}', then puts {var: f.__name__} into getters.
        '''
        fname = f.__name__
        START = 'get_'
        if not fname.startswith(START):
            raise AssertionError(f'expected f.__name__ to start with "{START}", but got {f.__name__!r}')
        name = fname[len(START):]
        getters[name] = fname
        return f

    @_param_getter
    def get_n(self, precision=3, *, units='raw', **kw):
        '''return dict of values to use for n0d for all fluids in self.ec.fluid.
        keys are n0d0, n0d1, etc. values are EppicParam objects, with old=n0d from input deck.
        precision is for converting to str in exponential form.

        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        values = dict()
        ec = self.ec
        with ec.using(units=units, **kw):
            assert ec.current_n_snap()==1, 'get_n expects self.ec.snap to be a single snap.'
            for fluid in ec.iter_fluid():
                old = fluid.get('n0d', None)
                key = f'n0d{int(fluid)}'
                n = ec('mean_n', item=True)
                values[key] = EppicParam(n, old=old, fmt=f'{{:.{precision}e}}')
        return values

    @_param_getter
    def get_v0(self, precision=3, *, units='raw', **kw):
        '''return dict of values to use for vx0d, vy0d, vz0d for all fluids in self.ec.fluid,
        for component in ec.component (e.g. use ec.component=('x', 'y') to ignore 'z')
        keys are vx0d0, vx0d1, etc. values are EppicParam objects, with old=vx0d from input deck.
        precision is for converting to str in exponential form.

        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        values = dict()
        ec = self.ec
        with ec.using(units=units, **kw):
            assert ec.current_n_snap()==1, 'get_v0 expects self.ec.snap to be a single snap.'
            for fluid in ec.iter_fluid():
                for x in ec.iter_component():
                    old = fluid.get(f'v{x}0d', None)
                    key = f'v{x}0d{int(fluid)}'
                    u = ec('mean_u', item=True)
                    values[key] = EppicParam(u, old=old, fmt=f'{{:.{precision}e}}')
        return values

    @_param_getter
    def get_vtherm(self, precision=3, *, units='raw', **kw):
        '''return dict of values to use for vxthd, vythd, vzthd for all fluids in self.ec.fluid,
        for component in ec.component (e.g. use ec.component=('x', 'y') to ignore 'z')
        keys are vxthd0, vxthd1, etc. values are EppicParam objects, with old=vxthd from input deck.
        precision is for converting to str in exponential form.

        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        values = dict()
        ec = self.ec
        with ec.using(units=units, **kw):
            assert ec.current_n_snap()==1, 'get_vtherm expects self.ec.snap to be a single snap.'
            for fluid in ec.iter_fluid():
                for x in ec.iter_component():
                    old = fluid.get(f'v{x}thd', None)
                    key = f'v{x}thd{int(fluid)}'
                    vtherm = ec('mean_vtherm', item=True)
                    values[key] = EppicParam(vtherm, old=old, fmt=f'{{:.{precision}e}}')
        return values

    @_param_getter
    def get_vtherm_n(self, precision=3, *, units='raw', **kw):
        '''return dict of {'vth_neutral': EppicParam(value to use for vth_neutral, old=value from input deck)}.

        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        ec = self.ec
        with ec.using(units=units, **kw):
            assert ec.current_n_jfluid() == 1, 'get_vtherm_n expects exactly one jfluid.'
            assert all(jfluid.is_neutral() for jfluid in ec.iter_jfluid()), 'get_vtherm_n expects neutral jfluid.'
            vtherm_n = ec.getj('mean_vtherm', item=True)
        old = ec.input_deck.get('vth_neutral', None)
        return {'vth_neutral': EppicParam(vtherm_n, old=old, fmt=f'{{:.{precision}e}}')}

    @_param_getter
    @format_docstring(default_ldebye_safety=DEFAULTS.EPPIC.DSPACE_SAFETY)
    def get_dx(self, precision=1, *, ldebye_safety=UNSET, units='raw', **kw):
        '''return dict of values to use for dx, dy, dz.
        keys are dx, dy, dz. values are EppicParam objects, with old=dx, dy, dz from input deck.
        precision is for converting to str in exponential form.

        ldebye_safety: UNSET or number
            dx, dy, dz = ldebye_safety * debye length of electrons.
            UNSET --> use DEFAULTS.EPPIC.DSPACE_SAFETY (default: {default_ldebye_safety})
        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        if ldebye_safety is UNSET: ldebye_safety = DEFAULTS.EPPIC.DSPACE_SAFETY
        values = dict()
        ec = self.ec
        with ec.using(units=units, **kw):
            assert ec.current_n_snap()==1, 'get_dx expects self.ec.snap to be a single snap.'
            KEYS = ('dx', 'dy', 'dz')
            old = {key: ec.input_deck.get(key, None) for key in KEYS}
            electron = ec.fluids.get_electron()
            ldebe = ec('mean_ldebye', fluid=electron, item=True)
            dx = ldebye_safety * ldebe
            values.update({key: EppicParam(dx, old=old[key], fmt=f'{{:.{precision}e}}') for key in KEYS})
        return values

    @_param_getter
    @format_docstring(default_dt_safety=DEFAULTS.EPPIC.DT_SAFETY)
    def get_dt(self, precision=1, *, dt_safety=UNSET, units='raw', **kw):
        '''return dict of {{'dt': EppicParam(value to use for dt, old=dt from input deck)}}.
        precision is for converting to str in exponential form.

        dt_safety: UNSET or number
            dt = dt_safety * minimum timescale.
            UNSET --> use DEFAULTS.EPPIC.DT_SAFETY (default: {default_dt_safety})
        units, kw:
            passed to self.ec.using(...) while using this method.
        '''
        if dt_safety is UNSET: dt_safety = DEFAULTS.EPPIC.DT_SAFETY
        with self.ec.using(**kw):
            timescales = self.timescales(units=units)
        dt = dt_safety * timescales.min()
        old = self.ec.input_deck.get('dt', None)
        return {'dt': EppicParam(dt, old=old, fmt=f'{{:.{precision}e}}')}

    def timescales(self, *, units='raw', **kw_init):
        '''returns an EppicTimescales object, to help with understanding the relevant timescales.
        kwargs are pased to EppicTimescales(...)
        '''
        return EppicTimescales(self.ec, units=units, **kw_init)

    # # CHECKERS # #
    def _param_checker(f, checkers=CHECKERS):
        '''return f after putting f.__name__ into checkers
        assert f.__name__ looks like 'check_{var}', then puts {var: f.__name__} into checkers.
        '''
        fname = f.__name__
        START = 'check_'
        if not fname.startswith(START):
            raise AssertionError(f'expected f.__name__ to start with "{START}", but got {f.__name__!r}')
        name = fname[len(START):]
        checkers[name] = fname
        return f

    def failed_check(self, msg):
        '''what to do when a check fails. depends on self.checks_mode.
            'warn' --> print warning message.
            'crash' --> raise AssertionError.
            True --> equivalent to 'warn'
            False --> do nothing.
        '''
        if self.checks_mode == 'warn' or self.checks_mode is True:
            print(f'WARNING: {msg}')
        elif self.checks_mode == 'crash':
            raise AssertionError(msg)
        elif self.checks_mode is False:
            pass  # do nothing.
        else:
            raise InputError(f'invalid checks_mode={self.checks_mode!r}. Must be True, False, "warn", or "crash".')

    @_param_checker
    def check_dt(self):
        '''checks that input_deck dt is smaller than or equal to the value from self.get_dt().
        returns self.get_dt(units='raw'), since it tells the suggested and old values for dt.
        '''
        dt = self.get_dt(units='raw')['dt']  # EppicParam for dt. 'raw' units to match input deck.
        value = dt.eval_str_value()
        old = dt.eval_str_old()
        if old is not None and old > value:
            self.failed_check(f'input_deck dt = {dt.str_old()} is too large! suggest dt <= {dt.str_value()}.')
        return dt

    @_param_checker
    def check_dx(self):
        '''checks that input_deck dx is smaller than or equal to the value from self.get_dx().
        returns self.get_dx(units='raw'), since it tells the suggested and old values for dx.
        '''
        dspace = self.get_dx(units='raw')  # dict of EppicParam for dx, dy, dz. 'raw' units to match input deck.
        for key, dx in dspace.items():
            value = dx.eval_str_value()
            old = dx.eval_str_old()
            if old is not None and old > value:
                self.failed_check(f'input_deck {key} = {dx.str_old()} is too large! suggest {key} <= {dx.str_value()}.')
        return dspace

    @_param_checker
    @format_docstring(default_rosenberg_safety=DEFAULTS.EPPIC.ROSENBERG_SAFETY)
    def check_rosenberg_qn(self, *, rosenberg_safety=UNSET):
        '''check the rosenberg criterion for quasineutrality, for each fluid: (nusn / wplasma)^2 << 1.
        
        rosenberg_safety: UNSET or number
            the check passes iff (nusn / wplasma)^2 <= rosenberg_safety.
            UNSET --> use DEFAULTS.EPPIC.ROSENBERG_SAFETY (default: {default_rosenberg_safety})

        return xarray of (mean(nusn / wplasma))^2 for each fluid
        '''
        ec = self.ec
        assert ec.current_n_snap()==1, 'check_rosenberg_qn expects self.ec.snap to be a single snap.'
        if rosenberg_safety is UNSET: rosenberg_safety = DEFAULTS.EPPIC.ROSENBERG_SAFETY
        rosenberg_qn = ec('max_rosenberg_qn')  # max <--> checking the "worst case".
        for fluid, val in ec.take_fluid(rosenberg_qn, as_dict=True).items():
            v = val.item()  # item() <--> expecting a single value here.
            if v > rosenberg_safety:
                errmsg = (f'Failed rosenberg_qn, for fluid {fluid}.'
                          f'\n    Expected (nusn / wplasma)^2 << 1, but got {v:.3g}, which is too large.'
                          f'\n    (Larger than {rosenberg_safety:.3g}, from DEFAULTS.EPPIC.ROSENBERG_SAFETY).')
                self.failed_check(errmsg)
        return rosenberg_qn


''' --------------------- EppicTimescales --------------------- '''

class EppicTimescales():
    '''helps with calculating dt based on physical timescales for eppic.

    ec: EppicCalculator
        used for calculations of values.
    units: str, default 'raw'
        units system for output values. Probably 'raw' or 'si'.
    item: bool, default True
        True --> convert all values to scalars (e.g., python floats).
        False --> keep values as xarrays.
    calc: bool, default True
        whether to calc all values when initializing, via self.calc().
    precision: None or int
        precision to use when converting to strings.
        None --> use type(self).precision (default 1)

    self.tt will be a dict like {scalename: {key: value}},
        where key will be:
            fluid, for timescales associated with a specific fluid;
            'global', for global timescales;
            'min', for timescales which are the minimum (e.g., across fluids).
    '''
    precision = 1  # digits of precision for exponential values when converting to strings.

    def __init__(self, ec, *, units='raw', item=True, calc=True, precision=None):
        self.tt = dict()
        self.ec = ec
        self.units = units
        self.item = item
        if precision is not None:
            self.precision = precision
        if calc:
            self.calc()

    # # # CALCULATING TIMESCALES # # #
    def values(self):
        '''returns a list of all timescale from self, in no particular order.'''
        return [value for d in self.tt.values() for value in d.values()]

    def min(self):
        '''returns the minimum timescale from within self.'''
        return min(self.values())

    def calc(self):
        '''calculate all timescales, updating self.tt. returns self.'''
        self.calc_timescales()
        self.calc_subcycling()
        self.calc_mins()
        return self

    def calc_timescales(self):
        '''calculate timescales for fluids, updating self.tt. returns self.
        calculations are in units system defined by self.units (probably 'raw' or 'si').

        (the actual math for timescales is in TimescalesLoader, in timescales.py)
        '''
        tt = self.tt
        ec = self.ec
        item = self.item
        units = self.units
        with ec.using(units=units):
            if item and not ec.current_n_snap()==1:
                errmsg = 'calc_timescales expectes self.ec.snap to be a single value. (when self.item=True).'
                raise DimensionError(errmsg)
            self.snap = ec.snap   # save the current snap, for display purposes.
            # plasma frequency
            d = tt.setdefault('wplasma', dict())
            for fluid in ec.iter_fluid():
                d[fluid] = ec('min_timescale_wplasma', item=item)  # min across space.
            # gryofrequency
            d = tt.setdefault('gyrof', dict())
            for fluid in ec.iter_fluid():
                d[fluid] = ec('min_timescale_gyrof', item=item)
            # collision frequency
            d = tt.setdefault('nusn', dict())
            for fluid in ec.iter_fluid():
                d[fluid] = ec('min_timescale_nusn', item=item)
            # thermal velocity (over dx). (Note, csound is basically vtherm.)
            d = tt.setdefault('vtherm', dict())
            for fluid in ec.iter_fluid():
                d[fluid] = ec('min_timescale_vtherm', item=item)
            # |E| / |B| speed (over dx)
            d = tt.setdefault('E/B', dict())
            d['global'] = ec('min_timescale_EBspeed', item=item)
        return tt

    def calc_subcycling(self):
        '''calculate timescales incorporating subcycling of fluids, updating self.tt. returns self.
        Equivalent to dividing every fluid timescale by `subcycle` for that fluid.
        scalenames will be old scalename with '_subcycle' appended.

        Does not remove old scalenames.
        '''
        tt = self.tt
        ec = self.ec
        subcycle = self.ec('subcycle')
        for scalename, d in list(tt.items()):
            if scalename.endswith('_subcycle'):
                continue  # don't subcycle the same thing twice by accident...
            ttsub = tt.setdefault(f'{scalename}_subcycle', dict())
            for fluid, value in list(d.items()):
                if fluid == 'min':
                    continue
                elif fluid == 'global':
                    ttsub[fluid] = value
                else:
                    ttsub[fluid] = value / subcycle.sel(fluid=np.array(fluid)).item()
        return self

    def calc_mins(self):
        '''calculate 'min' for each scalename, updating self.tt. returns self.'''
        tt = self.tt
        for scalename, d in list(tt.items()):
            d['min'] = min(d.values())
        return self

    # # # TO PANDAS # # #
    def to_pandas(self):
        '''return pandas DataFrame of timescales.
        'min' and 'global' columns will appear before other columns.
        '''
        result = pd.DataFrame(self.tt).T  # .T --> header will be fluid names.
        # re-order
        header = result.columns.to_list()
        order = []
        if 'min' in header:
            order.append('min')
        if 'global' in header:
            order.append('global')
        order.extend([h for h in header if h not in order])
        result = result[order]
        return result

    # # # DISPLAY # # #
    def to_pandas_display(self):
        '''return self.to_pandas(), but prettier, for display purposes:
        - replaces NaNs with '---'.
        - converts numbers to exponential form using self.precision.
        - put a helpful caption, including self.units, self.ec.snap, and self.ec.input_deck['dt']
        - highlights the minimum value in each row (but not in the 'min' column)
        - highlights the minimum value in the 'min' column (in a different color ^)
        - puts a horizontal border above the first row whose name ends with '_subcycle'
        '''
        pdvals = self.to_pandas()
        result = pdvals.copy()
        # replace NaNs with '---'
        result = result.fillna('---')
        # convert numbers to exponential form
        def val_to_str(x):
            return x if isinstance(x, str) else f'{x:.{self.precision}e}'
        result = result.map(val_to_str)
        # # STYLER -- must (must occur after all data is editted) # #
        # add caption
        NEWLINE = '\n<br>'  # newline; works in html or not.
        caption = f'timescales [units={self.units!r}], at snap = {getattr(self, "snap", None)!r}'
        run_title = self.ec.input_deck.get('title', None)
        if run_title is not None:
            caption += f'{NEWLINE}(run title = {run_title!r})'
        dt_raw = self.ec.input_deck.get('dt', None)
        if dt_raw is not None:
            dt = dt_raw * self.ec.u('t', self.units, convert_from='raw')
            caption += f'{NEWLINE}(dt from input deck = {dt} [{self.units!r}])'
            dtmin = self.min()
            if dt > dtmin:
                caption += f'{NEWLINE}>>>> WARNING: dt > min timescale ({val_to_str(dtmin)}) <<<<'
        #caption += f'{NEWLINE}(use {type(self).__name__}.to_pandas() to get pandas DataFrame.)'
        result = result.style.set_caption(caption)
        # highlight minimum value in each row
        def highlight_min(row):
            '''highlight the minimum value in this row.'''
            minval = row.iloc[0]  # min should be first column..
            result = ['']
            for v in row.iloc[1:]:
                if v == minval:
                    result.append('background-color: #abf7b1')  # light green
                else:  # not the min
                    result.append('')
            return result
        result = result.apply(highlight_min, axis=1)
        # highlight minimum value in the 'min' column
        def highlight_min_min(col):
            '''highlight the minimum value in the 'min' column.'''
            minval = val_to_str(np.nanmin(pdvals))
            result = []
            for v in col:
                if v == minval:
                    result.append('background-color: #5ced73')  # slightly darker light green
                else:   
                    result.append('')
            return result
        result = result.apply(highlight_min_min, axis=0, subset=['min'])
        # put a border above the first row whose name ends with '_subcycle'
        # (but only if _subcycle isn't the first row)
        rownames = result.index.to_list()
        subcycle_rows = [i for i, name in enumerate(rownames) if name.endswith('_subcycle')]
        if len(subcycle_rows) > 0:
            result.set_properties(subset=(result.index[subcycle_rows[0]],),
                                  **{'border-top': '2px solid black'})
        return result

    def _repr_html_(self):
        '''display hook for jupyter notebook.'''
        return self.to_pandas_display()._repr_html_()

    def __repr__(self):
        return f'{type(self).__name__}(tt=dict with keys {list(self.tt.keys())})'



"""
File Purpose: MainDimensionsHaver

Since it may include multiple axes, and many values along each axis,
the code architecture for MainDimensionsHaver is different than for the other DimensionHavers.

main dimensions are also NOT included in the regular list of dimensions,
to avoid iterating through each point in space when iterating through dimension values.

"dimpoints" in PlasmaCalcs refers to a point in the space of other dimensions,
i.e. a set of values for (some or all of) the other dimensions, but not maindims,
e.g. (snap=0, fluid=0, jfluid=1)
"""
import numpy as np
import xarray as xr

from .dimension_tools import DimensionHaver, MultiSlices, _paramdocs_multi_slices
from .snaps import INPUT_SNAP
from ..errors import LoadingNotImplementedError
from ..defaults import DEFAULTS
from ..tools import (
    format_docstring,
    simple_property, alias_child,
    product,
    xarray_isel, is_iterable_dim,
    interprets_fractional_indexing, is_iterable,
)


class UnsettableDict(dict):
    '''a dict whose values cannot be set. raise ValueError if attempting to set any items.'''
    def __init__(self, *args, errmsg='cannot set values in UnsettableDict', **kw):
        self.errmsg = errmsg
        super().__init__(*args, **kw)
    def __setitem__(self, key, value):
        raise ValueError(self.errmsg)
    def update(self, *args, **kw):
        raise ValueError(self.errmsg)


class MainDimensionsHaver(DimensionHaver):
    '''All the dimensions which remain after indexing by all other DimensionHavers.
    E.g., for simulation output from a 1024 x 512 x 512 simulation (in x, y, z),
        for a given fluid, jfluid, component, and snapshot,
        the array would be (1024, 512, 512).
    This is probably the shape of (most) arrays as stored in memory.

    use self.slices and self.slicing to control slicing behavior

    Subclasses should implement:
        maindims: listlike of strs
            tells main dimensions. Will be accessed as instance.maindims.
            Best to use a property if maindims depends on class.
        get_maindims_coords: method
            self.get_maindims_coords() should return dict of {dim: coords} for dims in self.maindims.
            it should also slice coords appropriately, according to self.slices;
                for this task, one helpful method might be self._apply_maindims_slices_to_dict().
    '''
    # # # BEHAVIOR ATTRS # # #

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['multi_slices_ndim', 'multi_slices_ikeep'] + any values from super().
        '''
        return ['multi_slices_ndim', 'multi_slices_ikeep'] + super()._extra_kw_for_quantity_loader_call

    # # # MAINDIMS -- SUBCLASS SHOULD IMPLEMENT # # #
    maindims = NotImplemented   # listlike of main dimensions, e.g. ['x', 'y', 'z']

    def get_maindims_coords(self):
        '''return dict of {dim: coords} for all dimensions in self.main_dims.
        E.g., {'x': xcoords, 'y': ycoords, 'z': zcoords}, if main dimensions are x, y, z.
        coords will each be sliced using the appropriate slices from self.slices.
        '''
        raise LoadingNotImplementedError(f'{type(self).__name__}.get_maindims_coords')

    # # # MAINDIMS -- IMPLEMENTED HERE # # #
    @property
    def maindims_shape(self):
        '''tuple of (len(self.get_maindims_coords()[dim]) for dim in self.maindims).
        Note, this should be sensitive to changes in self.slices. See also: self.maindims_full_shape.
        '''
        coords = self.get_maindims_coords()
        result = []
        for dim in self.maindims:
            try:
                L = len(coords[dim])
            except TypeError:
                L = 1
            result.append(L)
        return tuple(result)

    @property
    def maindims_size(self):
        '''product of terms in self.maindims_shape.
        Note, this should be sensitive to changes in self.slices. See also: self.maindims_full_size.
        '''
        return product(self.maindims_shape)

    @property
    def maindims_sizes(self):
        '''dict of {dim: size of dim} for dim in self.maindims.
        Note, this should be sensitive to changes in self.slices. See also: self.maindims_full_sizes.
        '''
        shape = self.maindims_shape
        return {dim: shape[i] for i, dim in enumerate(self.maindims)}

    @property
    def maindims_full_shape(self):
        '''self.maindims_shape when self.slices=None'''
        with self.using(slices=None):
            return self.maindims_shape

    @property
    def maindims_full_size(self):
        '''self.maindims_size when self.slices=None'''
        with self.using(slices=None):
            return self.maindims_size

    @property
    def maindims_full_sizes(self):
        '''self.maindims_sizes when self.slices=None'''
        with self.using(slices=None):
            return self.maindims_sizes

    cls_behavior_attrs.register('maindims_means', default=False)
    maindims_means = simple_property('_maindims_means', default=False,
            doc='''whether to immediately take means across maindims when loading arrays. (default False.)
            True --> treat data across maindims as if it were the mean values, only.
                    Caution: this is different from taking means after doing calculations;
                    e.g., with maindims_means = True, 'n*T' --> mean(n)*mean(T), not mean(n*T).''')

    # # # SLICES # # #
    slicing = simple_property('_slicing', default=True,
            doc='''whether to slice maindims when loading arrays & during get_maindims_coords.
            if False, self.slices will return an empty dict.''')

    cls_behavior_attrs.register('slices', default={})
    @property
    def slices(self):
        '''slices for maindims when loading arrays & during get_maindims_coords.
        E.g. slices = dict(x=slice(0,50), y=7)
            --> slice arrays along x & y, taking the first 50 x values, and only the 7th y value.

        Notes:
            - only applies slices along arrays which actually contain the related coordinates,
                e.g. if z=10 appears in slice but loading an array with only x & y, won't apply z=10 slice.
            - supports fractional indexing, as per interprets_fractional_indexing.
                Non-integer values between -1 and 1 can be used to infer to a fraction of the dimension length,
                with negative values referring to a distance from the end, just like with integer indexing.
                Example: dict(x=slice(-0.3, None, 0.01), y=0.8), where x and y each have length 1000
                    --> equivalent to dict(x=slice(-300, None, 10), y=800).

        if self.slicing is False, self.slices will give an empty dict and cannot be set to any value!
            however, the old value of self.slices will be remembered in case slicing is set to True later.
        '''
        if not self.slicing:
            return UnsettableDict(errmsg='cannot set slices when slicing is False.')
        try:
            return self._slices
        except AttributeError:
            self._slices = dict()
            return self._slices
    @slices.setter
    def slices(self, value):
        if not self.slicing:
            raise ValueError('cannot set slices when slicing is False.')
        if value is None:
            value = dict()
        self._slices = value
    @slices.deleter
    def slices(self):
        del self._slices

    def _apply_maindims_slices_to_dict(self, d):
        '''slice entries in dict using self.slices. returns d. (Note: d will be directly altered.)'''
        for key, s_ in self.slices.items():
            if key in d:
                slice_ = interprets_fractional_indexing(s_, len(d[key]))
                d[key] = d[key][slice_]
        return d

    def standardized_slices(self):
        '''returns a copy of self.slices, but calling interprets_fractional_indexing on all slices,
        using lengths from self.maindims_full_sizes.
        '''
        maindims_sizes = self.maindims_full_sizes
        result = {}
        for key, s_ in self.slices.items():
            result[key] = interprets_fractional_indexing(s_, maindims_sizes.get(key, None))
        return result

    def slicestr(self, *, sep=', ', keep_None=False):
        '''string representation of self.slices, for use in filenames, titles, etc.
        comma-separated, alphabetized, ignoring slice(None).
        Supports single-indexes (e.g. x=5), slices (e.g. y=slice(0, 4)),
            and fractional indexing (e.g. z=slice(0, 0.5, 0.01)),
            though fractional indexing will be converted to ints.

        sep: str, separator between slices
        keep_None: bool, whether to keep slices with value None in the string.
        '''
        slices = self.slices
        order = sorted(slices.keys())
        results = []
        maindims_sizes = None
        for key in order:
            ss = slices[key]
            if ss == slice(None):
                if keep_None:
                    results.append(f'{key}=None')
                else:
                    continue
            elif isinstance(ss, int):
                results.append(f'{key}={ss}')
                continue
            # else
            if maindims_sizes is None:  # first time getting sizes
                with self.using(slices=None):
                    maindims_sizes = self.maindims_sizes
            ss = interprets_fractional_indexing(ss, L=maindims_sizes.get(key, None))
            if isinstance(ss, slice):
                if ss.start is None and ss.step is None:
                    results.append(f'{key}=slice({ss.stop})')
                elif ss.step is None:
                    results.append(f'{key}=slice({ss.start},{ss.stop})')
                else:
                    results.append(f'{key}=slice({ss.start},{ss.stop},{ss.step})')
            elif is_iterable(ss):
                raise InputError(f'cannot produce slicestr with iterable self.slices[{key!r}]={ss}')
            else:  # probably an int
                results.append(f'{key}={slice_}')
        return sep.join(results)

    def title_with_slices(self, *, sep=', ', keep_None=False):
        '''return self.title with slicestr appended (after sep), if slicestr is not empty.
        see self.slicestr() for more details.
        '''
        slicestr = self.slicestr(sep=sep, keep_None=keep_None)
        title = self.title
        return title if slicestr == '' else f'{title}{sep}{slicestr}'

    # # # MULTI-SLICES # # #
    cls_behavior_attrs.register('multi_slices', getdefault=lambda ql: MultiSlices(dims=ql.maindims))
    @property
    @format_docstring(**_paramdocs_multi_slices, sub_ntab=2)
    def multi_slices(self):
        '''dict of {{key: slices dict}}.
        When getting any vars across maindims, make a Dataset by applying each of these, separately.
        If len(multi_slices)>0 then ignore self.slices.

        Can also provide special keys 'ndim' and/or 'ikeep' to create special slices:
            Example: if self.maindims=['x', 'y', 'z'], then self.multi_slices = dict(ndim=2, ikeep=0)
                is equivalent to: self.multi_slices = dict(x_y=dict(z=0), x_z=dict(y=0), y_z=dict(x=0))
            Details:
                ndim: {ndim}
                ikeep: {ikeep}

            Can also set these as attributes of self.multi_slices to achieve the same effect.
            E.g. self.multi_slices.ndim = 2
        '''
        if not self.slicing:
            return UnsettableDict(errmsg='cannot set multi_slices when slicing is False.')
        try:
            return self._multi_slices
        except AttributeError:
            self._multi_slices = self._new_multi_slices()
            return self._multi_slices
    @multi_slices.setter
    def multi_slices(self, value):
        '''set self.multi_slices. If value is not an instance of MultiSlices, convert it to one.'''
        if not self.slicing:
            raise ValueError('cannot set multi_slices when slicing is False.')
        if not isinstance(value, self._multi_slices_cls):
            if not isinstance(value, dict):
                errmsg = (f'multi_slices must be a dict or a {self._multi_slices} object,'
                          f' but got type(value)={type(value)}.')
                raise ValueError(errmsg)
            value = self._new_multi_slices(**value)  # dict but not MultiSlices
        self._multi_slices = value
    @multi_slices.deleter
    def multi_slices(self):
        del self._multi_slices

    _multi_slices_cls = MultiSlices  # cls used for self.multi_slices.

    def _new_multi_slices(self, **kw):
        '''called when self.multi_slices accessed but not yet set; create & return new MultiSlices.'''
        kw.setdefault('dims', self._multi_slices_dims_default)
        return self._multi_slices_cls(**kw)

    def _multi_slices_dims_default(self):
        '''function to get multi_slices dims, by default.
        Would love to use (lambda: self.maindims) instead of this function,
            however lambdas are not pickleable, thus incompatible with multiprocessing module.
        Providing this function allows the default multi_slices to still be pickleable.
        '''
        return self.maindims

    multi_slices_ndim = alias_child('multi_slices', 'ndim', doc=_paramdocs_multi_slices['ndim'])
    multi_slices_ikeep = alias_child('multi_slices', 'ikeep', doc=_paramdocs_multi_slices['ikeep'])

    # # # LOAD MAINDIMS VAR # # #
    def load_maindims_var(self, var, *args, u=None, assign_labels=True, **kw):
        '''return var, formatted as an xarray with proper details for PlasmaCalcs.
        loading var should give an array with self.maindims as dimensions.

        Also does these steps:
            1) assign maindims coords via self.assign_maindims_coords().
            2) slice array via self.slices.
            3) convert units, if u is not None
            4) set result.attrs['units'] = self.units
            5) if self.maindims_means: take mean of result, across all maindims.
            6) use result = self._maindims_postprocess_callback(result), if possible.

        u: None, value, or str
            units factor for the result.
            None --> don't do any units conversions.
            str --> multiply result by self.u(u)
            value --> multiply result by u
        assign_labels: bool
            whether to assign_maindims_coords and self.record_units.
            Recommend to always use True, unless using this function internally.
            (e.g. for load_maindims_var_across_dims, only use the first time, for efficiency.)
            IGNORED if self.maindims_means.

        Note:
            If load_direct(var) uses an override or gets from cache or self.setvars,
                skip steps 1,2,3,4
            ([TODO] Might need to reconsider this behavior?)
        Note:
            If self.multi_slices are provided, load_maindims_var for each slice,
            then combine results into an xarray.Dataset.
            if assign_labels=False, combine results into a dict instead.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # if multi_slices are provided, load_maindims_var for each slice then combine results.
        orig_multi_slices = self.multi_slices.get_slices()
        if len(orig_multi_slices) > 0:
            with self.using(multi_slices=dict()):  # don't recurse on multi_slices.
                results = dict()
                for key, mslice in orig_multi_slices.items():
                    with self.using(slices=mslice):
                        results[key] = self.load_maindims_var(var, *args, u=None,
                                                              assign_labels=assign_labels, **kw)
            if assign_labels:
                # convert to Dataset; be sure to keep any common attrs too!
                keys = list(results.keys())
                if len(keys) == 0:
                    common_attrs = None
                elif len(keys) == 1:
                    common_attrs = results[keys[0]].attrs.copy()
                else:
                    common_attrs = results[keys[0]].attrs.copy()
                    for key in keys[1:]:
                        for k, v in results[key].attrs.items():
                            if k in common_attrs and common_attrs[k] != v:
                                del common_attrs[k]
                return xr.Dataset(results, attrs=common_attrs)
            else:  # if not assigning labels, return as dict.
                return results
        # <-- multi_slices were not provided.
        # bookkeeping
        assigning_labels = (assign_labels or self.maindims_means)
        # [TODO](maybe) internally maybe write two different functions,
        #   one with assigning_labels=True, one with assigning_labels=False...
        # get value
        result = self.load_direct(var, *args, **kw)
        if getattr(self, '_load_direct_used_override', False) is not None:
            # either used an override, or got value from cache or self.setvars.
            # in this case, skip steps 1-4. (see docstring)
            pass
        elif getattr(self, 'snap', None) is INPUT_SNAP:
            # used the "override": load value based on eppic.i file.
            pass
        else:
            # got result directly from a file. Do steps 1-4. (see docstring).
             # (1) assign maindims coords
            if assigning_labels:
                result = self.assign_maindims_coords(result)
            # (2) slice array (unless already sliced it in load_direct)
            if not self._slice_maindims_in_load_direct:
                if assigning_labels:
                    result = self.slice_maindims(result)
                else:
                    result = self._slice_maindims_numpy(result)
            # (3) convert units
            if u is not None:
                if isinstance(u, str):
                    u = self.u(u)
                result = result * u
            # (4) set result.attrs['units']
            if assigning_labels:
                try:
                    self_record_units = self.record_units
                except AttributeError:
                    pass  # that's fine, we just won't record units.
                else:
                    result = self_record_units(result)
        # (5) take mean across maindims, if applicable.
        if self.maindims_means:
            # only take mean across maindims which actually appear in result.dims
            if hasattr(result, 'dims') and hasattr(result, 'mean'):
                dims = result.dims
                to_mean = [dim for dim in self.maindims if dim in dims]
                if len(to_mean) > 0:
                    result = result.mean(to_mean)
        # (6) postprocess with self._maindims_postprocess_callback, if applicable.
        try:
            callback = self._maindims_postprocess_callback
        except AttributeError:
            pass  # that's fine, just don't do any postprocessing.
        else:
            if callback is not None:
                result = callback(result)
        # finished. return result.
        return result

    _slice_maindims_in_load_direct = False  # if True, load_direct should slice maindims

    def _load_maindims_var_with_labels(self, var, *args, **kw):
        '''equivalent to self.load_maindims_var(var, *args, assign_labels=True, **kw)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.load_maindims_var(var, *args, assign_labels=True, **kw)

    def _load_maindims_var_without_labels(self, var, *args, **kw):
        '''equivalent to self.load_maindims_var(var, *args, assign_labels=False, **kw)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.load_maindims_var(var, *args, assign_labels=False, **kw)

    def assign_maindims_coords(self, array):
        '''assign maindims dims and coords, based on self.get_maindims_coords() with slicing=False.
        array must have same shape as implied by maindims and coords.
            if array is 0D, just return a 0D xr.DataArray.
        returns an xarray with proper details for PlasmaCalcs.

        This function creates a *new* xarray based on array, and maindims & coords are >0 dimensional.
        This is not like assign_{dim}_coord functions, which assign 0D coord to an existing xarray.
        '''
        ndim = np.ndim(array)
        if ndim == 0:
            return array if isinstance(array, xr.DataArray) else xr.DataArray(array)
        dims = self.maindims
        if self._slice_maindims_in_load_direct and self.slicing:
            coords = self.get_maindims_coords()
            # any non-iterable coords are not dims
            dims = [d for d in dims if is_iterable_dim(coords[d])]
        else:
            assert len(self.maindims) == ndim, f"expected {len(self.maindims)}D, but got ndim={np.ndim(array)}"
            with self.using(slicing=False):
                coords = self.get_maindims_coords()
        array = xr.DataArray(array, coords=coords, dims=dims)
        return array

    def _slice_maindims_numpy(self, array, *, h5=False):
        '''slice first len(maindims) dims of array, using self.slices.

        h5: bool
            whether 'array' might be an h5py dataset.
            if True, handle negative step in the intuitive way;
                i.e. slice with positive step, to select the expected points, then reverse order.
                (This is necessary because h5 datasets crash when sliced by negative step...)
        '''
        slices = self.slices
        if len(slices) == 0:
            return array
        slices = [slices.get(dim, slice(None)) for dim in self.maindims]
        # handle fractional indexing
        shape = array.shape
        for i, slice_ in enumerate(slices):
            slices[i] = interprets_fractional_indexing(slice_, shape[i])
        if h5:  # handle negative steps
            pre_slices = slices
            post_slices = [slice(None) for _ in slices]
            for i, (pre, post) in enumerate(zip(pre_slices, post_slices)):
                if isinstance(pre, slice) and (pre.step is not None) and (pre.step < 0):
                    raise NotImplementedError('[TODO] slicing with step<0, when h5=True')
        return array[tuple(slices)]

    def _slices_which_scalarize(self):
        '''return list of maindims which become scalars when hit by self.slices.'''
        slices = self.slices
        result = []
        for x in self.maindims:
            slicex = slices.get(x, None)
            if not ((slicex is None) or isinstance(slicex, slice) or is_iterable(slicex)):
                result.append(x)
        return result

    def slice_maindims(self, array, **kw_xarray_isel):
        '''slice maindims of array using self.slices. See help(type(self).slices) for more details.
        (if slices is an empty dict, return array, unchanged, without making a copy.)
        Only slice dims which actually appear in array.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        slices = self.slices
        if len(slices) == 0:
            return array
        array_dims = array.dims
        if len(array_dims) == 0:  # [EFF] quickcheck to handle 0D case
            return array
        to_slice = {k: v for k, v in slices.items() if k in array_dims}
        if len(to_slice) == 0:
            return array
        kw_xarray_isel.setdefault('promote_dims_if_needed', False)  # don't promote non-dimension coords here.
        return xarray_isel(array, to_slice, **kw_xarray_isel)  # <-- handles fractional indexing internally.

    # # # LOAD MAINDIMS VAR ACROSS DIMS # # #
    def load_maindims_var_across_dims(self, var, dims=None, *, skip=[], u=None, **kw):
        '''load maindims var across these dims. Use all dims from self.dimensions if dims is None.
        Only loads across the current value of these dims (e.g., self.fluid, not self.fluids).
            (Can set current value to multiple values e.g. self.component = ('x', 'y').)

        u: None, value, or str
            units factor for the result.
            None --> don't do any units conversions.
            str --> multiply result by self.u(u)
            value --> multiply result by u
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        _shift_special = {'snap': [INPUT_SNAP]}  # [TODO] allow subclass to alter... don't hard-code it here.
        return self.load_across_dims(loader=self._load_maindims_var_without_labels, var=var,
                                     loader0=self._load_maindims_var_with_labels,
                                     dims=dims, skip=skip, u=u,
                                     _shift_special=_shift_special, **kw)



"""
File Purpose: The main calculator class for multifluid analysis from single-fluid mhd,
for hookups to inherit from.
"""

from .mhd_multifluid_bases import MhdMultifluidBasesLoader
from .mhd_multifluid_densities import MhdMultifluidDensityLoader
from .mhd_multifluid_ionization import MhdMultifluidIonizationLoader
from ..mhd_calculator import MhdCalculator
from ...plasma_calculator import MultifluidPlasmaCalculator

class MhdMultifluidCalculator(MhdMultifluidDensityLoader, MhdMultifluidIonizationLoader,
                              MhdMultifluidBasesLoader, MultifluidPlasmaCalculator,
                              MhdCalculator):
    '''class for multi-fluid analysis of single-fluid MHD outputs.

    set self.fluid=SINGLE_FLUID to get single-fluid values,
        otherwise will get inferred multifluid values.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    # parent class ordering notes:
    # - MhdMultifluidIonizationLoader must go before MhdMultifluidBasesLoader,
    #     because MhdMultifluidBasesLoader parent BasesLoader defines get_ionfrac too;
    #     this affects known_var results... (maybe it's a bug in known_var code?)
    #     with this ordering, KNOWN_VARS['ionfrac'].cls_where_defined is MhdMultifluidIonizationLoader;
    #     without this ordering, it is BasesLoader instead, which gives wrong deps.

    @property
    def fluids_elements(self):
        '''ElementList of unique elements found in any of self.fluids.
        To alter self.elements, adjust self.fluids; will infer new self.elements automatically.
        '''
        return self.fluids.unique_elements(istart=0)

    def init_fluids(self):
        '''initialize self.fluid, fluids, jfluid, and jfluids
        [Not implemented by MhdMultifluidCalculator; subclass should implement].
        '''
        raise NotImplementedError(f'{type(self).__name__}.init_fluids()')

    def use_mix_heavy_ions(self, m_tol=0.1, *, m_min=5, m_mean_mode='simple', **kw_mix_heavy_ions):
        '''self.init_fluids() then set self.fluids = self.fluids.mix_heavy_ions(...). returns self.fluids.
        See self.fluids.mix_heavy_ions for more details.

        m_tol: number
            maximum allowed relative mass deviation from mean within a group (an IonMixture).
            within each group, all ions have |m_ion - mean(m)| < m_tol * mean(m).
            E.g. 0.1 --> all ions in each group are within 10% of that group's mean mass.
        m_min: number
            minimum mass (in amu, i.e. m_H ~= 1) for an ion to be considered "heavy".
            ions with m < m_min are ignored.
        m_mean_mode: str
            mode for calculating mean mass of each IonMixture.
            'simple' (default) or 'density' (weighted by number density).
        '''
        self.init_fluids()
        mix = self.fluids.mix_heavy_ions(m_tol=m_tol, m_min=m_min, m_mean_mode=m_mean_mode, **kw_mix_heavy_ions)
        self.fluids = mix
        return self.fluids


    # # # COLLISIONS # # #
    # aliases to check during set_collisions_crosstab_defaults
    #    (which gets called the first time self.collisions_cross_mapping is accessed).
    # override from super() to avoid 'H' and 'He' aliases to avoid ambiguity with Element fluids.
    _COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES = \
        MultifluidPlasmaCalculator._COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES.copy()
    _COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES.update({
        'H_I'   : ['H_I',    'H I'],  # excludes 'H', since 'H' might be an Element.
        'He_I'  : ['He_I',   'He I'], # excludes 'He', since 'He' might be an Element.
        })

    # note: lots of other functionality inherited from parents.



"""
File Purpose: EbysusEfieldLoader
"""

import xarray as xr

from ...quantities import QuantityLoader
from ...tools import UNSET, xarray_sum

class EbysusEfieldLoader(QuantityLoader):
    '''calculate electric field from ebysus'''

    @known_var(dims=['snap', 'component'])
    def get_E_helita(self):
        '''electric field. (directly from helita)'''
        return self.load_maindims_var_across_dims('ef', dims=['snap', 'component'])

    @known_var(deps=['u', 'B'], ignores_dims=['fluid'])
    def get_E_uxB(self, *, _u_e=UNSET):
        '''electric field u x B contribution. E_uxB == -1 * u_electron cross B

        [EFF] for efficiency, can provide full u_e, if already known. (provide as _u_e)
                (requires full ue, not just one component, since will do ue cross B.)
        '''
        result = UNSET
        # in case appropriate _u_e was provided:
        if _u_e is not UNSET:
            uecomps = _u_e.coords.get('component', [])
            if not xr.core.utils.is_scalar(uecomps) and len(uecomps) == 3:
                B = self('B')
                result = self.cross_product(B, _u_e, components=self.component)
        # in case appropriate _u_e was not provided:
        if result is UNSET:
            result = self('B_cross_u', fluid=self.fluids.get_electron())
        return result.drop_vars('fluid')  # drop the fluid coordinate because result doesn't care about it.

    @known_var(deps=['grad_P', 'n', 'q'], ignores_dims=['fluid'])
    def get_E_bat(self, *, _u_e=None):
        '''electric field battery contribution. E_bat == grad(P_e) / (ne qe)'''
        with self.using(fluid=self.fluids.get_electron()):
            grad_Pe = self('grad_P')
            ne = self('n')
            qe = self('q')
        result = grad_Pe / (ne * qe)
        return result.drop_vars('fluid')  # drop the fluid coordinate because result doesn't care about it.

    @known_var(deps=['nu_ij', 'u', 'q'], ignores_dims=['fluid'])
    def get_E_momj(self, *, _u_e=UNSET):
        '''electric field momentum contribution due to jfluid.
        E_momj == -1 * me ne nu_ej * (uj - ue) / (ne qe)

        [EFF] for efficiency, can provide u_e if already known. (provide as _u_e)
            (only used if _u_e.coords['component'] has the exact same values as self.component.)
        '''
        with self.using(fluid=self.fluids.get_electron()):
            nuej = self('nusj')
            #ne = self('n')  # ne cancels out; don't need it.
            qe = self('q')
            me = self('m')
            uj = self.getj('u')
            ue = UNSET
            # in case appropriate _u_e was provided:
            if _u_e is not UNSET:
                uecomps = _u_e.coords.get('component', [])
                if not xr.core.utils.is_scalar(uecomps):
                    if set(c.item() for c in uecomps) == set(self.component):
                        ue = _u_e
            # in case appropriate _u_e was not provided:
            if ue is UNSET:
                ue = self('u')
        result = (me / qe) * nuej * (ue - uj)
        return result.drop_vars('fluid')  # drop the fluid coordinate because result doesn't care about it.

    @known_var(deps=['E_momj'], ignores_dims=['jfluid'])
    def get_E_mom(self, *, _u_e=UNSET):
        '''electric field momentum contribution. E_mom == -1 * sum_j me ne nu_ej * (uj - ue) / (ne qe)
        Equivalent: E_mom = sum(E_momj), where sum is taken across jfluids.

        [EFF] for efficiency, can provide full u_e, if already known. (provide as _u_e)
                (requires full ue, not just one component, since will do ue cross B.)
        '''
        # [EFF] skip jfluids which are known to have nu_ej == 0:
        electron = self.fluids.get_electron()
        jfluids = [fj for fj in self.jfluids if self('collision_type', fluid=electron, jfluid=fj).item() != '0']
        result = self('E_momj', jfluid=jfluids, _u_e=_u_e)
        return xarray_sum(result, dim='jfluid')
        # [EFF] below is significantly slower than above; maybe due to needing ue multiple times?
        # results = []
        # for fj in jfluids:
        #     results.append(self('E_momj', jfluid=fj))
        # return sum(results).drop_vars('jfluid')  # drop the jfluid coordinate because result doesn't care about it.

    @known_var(deps=['E_uxB', 'E_bat', 'E_mom'])
    def get_E_terms(self):
        '''xarray Dataset of the terms which contribute to E.
        Dataset with 'E_uxB', 'E_bat', and 'E_mom' as data vars.

        Suggestion: to quickly determine which E_terms are most important, consider using, e.g.:
            self('mod_E_terms').pc.stats(keep='variable')  # for mod across all components
            self('E_terms', component=0).pc.stats(keep='variable')  # for just component 0
        '''
        ue = self('u', fluid=self.fluids.get_electron())  # [EFF] calculate this only once.
        return self(['E_uxB', 'E_bat', 'E_mom'], _u_e=ue)

    @known_var(deps=['E_uxB', 'E_bat', 'E_mom'])
    def get_E(self):
        '''electric field. E = E_uxB + E_bat + E_mom'''
        # [EFF] (2024/03/16) ~20% faster than helita when doing all three components, thanks to _u_e.
        #     However, ~70% slower than helita when doing only 1 component.
        #     (tested with a 512x512 grid for ~10 snapshots.)
        # accuracy check: seems to be reasonably similar at dd.stagger_kind='first' or 'fifth'.
        # [TODO] further efficiency improvements?
        ue = self('u', fluid=self.fluids.get_electron())  # [EFF] calculate this only once.
        return self('E_uxB', _u_e=ue) + self('E_bat') + self('E_mom', _u_e=ue)



"""
File Purpose: reading runtime info for EppicCalculator

this is the timing information from the timers.dat file.
"""
import numpy as np
import xarray as xr

from .eppic_io_tools import read_timers_dat, eppic_clock_times_here
from ...quantities import QuantityLoader
from ...tools import (
    alias,
    find_jobfiles,
    xarray_promote_dim, xarray_sum,
)

class EppicRuntimeInfoLoader(QuantityLoader):
    '''runtime info for EppicCalculator.'''

    # # # FROM LOG FILE # # #

    def jobfiles(self):
        '''return list of all jobfiles within self.dirname directory.'''
        return find_jobfiles(self.dirname)

    logfiles = alias('jobfiles')

    def clock_times(self):
        '''return dict of clock times from this run. (From reading self.jobfiles.)
        Result can have keys:
            'start': datetime telling when the run started.
            'stepstart': datetime telling when the iterations started.
            'end': datetime telling when the run ended.
            'init_seconds': (stepstart - start) [seconds]
            'steps_seconds': (end - stepstart) [seconds]
            'total_seconds': (end - start) [seconds]
        If jobfile missing any info, relevant keys will not appear in result.
        '''
        return eppic_clock_times_here(self.dirname)

    @known_var
    def get_clock_times(self):
        '''dataset of clock times from this run. (From reading self.jobfiles.)
        Result can have keys:
            'start': datetime telling when the run started.
            'stepstart': datetime telling when the iterations started.
            'end': datetime telling when the run ended.
            'init_seconds': (stepstart - start) [seconds]
            'steps_seconds': (end - stepstart) [seconds]
            'total_seconds': (end - start) [seconds]
        If jobfile missing any info, relevant keys will not appear in result.
        '''
        return xr.Dataset(self.clock_times())

    @known_var(deps=['clock_times'])
    def get_init_seconds(self):
        '''time [in seconds] spent initializing the run. (between start and when steps start.)'''
        return self('clock_times')['init_seconds']

    @known_var(deps=['clock_times'])
    def get_steps_seconds(self):
        '''time [in seconds] spent doing timesteps. (total duration minus init_seconds.)'''
        return self('clock_times')['steps_seconds']

    @known_var(deps=['clock_times'], aliases=['total_time_seconds'])
    def get_total_seconds(self):
        '''duration [in seconds] spent to run the entire run.'''
        return self('clock_times')['total_seconds']


    # # # FROM TIMERS.DAT # # #

    def timers_dat(self, *, with_snaps=False, as_array=False):
        '''return timers.dat as an xarray.Dataset. (dimension will be named 'it')
        result will have the same units as timers.dat file.

        with_snaps: bool
            if True, attach snap & t coords and promote 'snap' to main dim.
            based on self.snaps (not self.snap)
        as_array: bool
            whether to use xarray.Dataset.to_array() to return a DataArray instead of Dataset.
            if True, vars from Dataset will be concatenated along the new dimension named 'timer'.
        '''
        result = read_timers_dat(self.dirname, as_array=as_array, # [TODO][EFF] use caching if this is slow.
                                fix_snaps=(self.snaps_from=='parallel'))
        if with_snaps:
            existing_snaps = self.existing_snaps()
            result = self.assign_snap_along(result, 'it', existing_snaps)
            result = xarray_promote_dim(result, 'snap')
            if len(existing_snaps) != len(self.snaps):  # some self.snaps refer to MISSING_SNAP
                # fill MISSING_SNAP points with NaN.
                result = result.reindex(snap=self.snaps, fill_value=self.snap_dim.NAN)
                # assign t coords to MISSING_SNAP points  # [TODO][EFF] do something more efficient?
                result = self.assign_snap_along(result, 'snap', self.snaps)
            result = result.drop_vars('it')
        return result

    @known_var(dims=['snap'])
    def get_timers(self):
        '''timers_dat info as an xarray.Dataset, at snaps in self.snap. see also: 'runtimes'.'''
        result = self.timers_dat(with_snaps=True, as_array=False)
        result = result.sel(snap=np.array(self.snap))
        return result

    @known_var(dims=['snap'])
    def get_runtimes(self):
        '''timers_dat info as an xarray.DataArray, at snaps in self.snap. see also: 'timers'.'''
        result = self.timers_dat(with_snaps=True, as_array=True)
        result = result.sel(snap=np.array(self.snap))
        return result

    @known_var(deps=['timers'])
    def get_run_time(self):
        '''Wall clock runtime for each snap. Same units as timers_dat info.'''
        return self('timers')['Wall Clock']

    @known_var(deps=['timers'])
    def get_write_time(self):
        '''time spent writing output files. Same units as timers_dat info.'''
        return self('timers')['output']

    @known_var(deps=['run_time', 'write_time'])
    def get_calc_time(self):
        '''Wall clock runtime, ignoring time spent writing output files. Same units as timers_dat info.'''
        return self('run_time') - self('write_time')

    @known_pattern(r'(write|calc)_time_frac', deps=['run_time', {0: '{group0}_time'}])
    def get_time_frac(self, var, *, _match=None):
        '''fraction of runtime spent on writing or calculating.
            var='write_time_frac' --> fraction of runtime spent writing output files.
            var='calc_time_frac' --> fraction of runtime spent calculating, ignoring write_time.
        '''
        here, = _match.groups()
        return self(f'{here}_time') / self('run_time')

    @known_pattern(r'(run|calc|write)_(timestep|dt)_cost(_f|_nosub|_fnosub|_nosubf)?',
            deps=['nit_since_prev', 'npd', {0: '{group0}_time'},
                {2: lambda groups: 'subcycle' if (groups[2] is not None and 'sub' in groups[2]) else []}])
    def get_timestep_cost_or_dt_cost(self, var, *, _match=None):
        '''total cpu time per simulated particle, per timestep (or per dt).
        
        time_cost = (runtime / timestep_or_dt) * (n_processors / total number of particles)
        total number of particles = n_processors * npart.
            Note: n_processors cancels out; time_cost = (runtime / timestep_or_dt) / npart
        timestep_or_dt = one timestep or one dt; see below.
        npart = number of simulated particles, in one processor. Depends on {settings}; see below.

        '{clock}_{time}_cost{settings}'
            E.g. 'run_timestep_cost', 'write_dt_cost_f', 'calc_dt_cost_nosubf'
            {clock} = 'run', 'calc', or 'write'
                tells which clock to use.
                'run' --> 'Wall clock' | 'calc' --> 'Wall Clock - output' | 'write' --> 'output'
            {time} = 'timestep' or 'dt'
                'timestep' --> report result as cost per timestep, regardless of dt.
                'dt'       --> report result as cost per dt (converted to SI units).
            {settings} = '', '_f', '_nosub', '_fnosub', or '_nosubf'
                tells whether to return a separate value for each fluid, and whether to account for subcycling.
                    ''        --> single value. account for subcycling.     npart = self('npd/subcycle').sum('fluid')
                    '_nosub'  --> single value. ignore subcycling.          npart = self('npd').sum('fluid')
                    '_f'      --> per-fluid values. account for subcycling. npart = self('npd/subcycle')
                    '_fnosub' --> per-fluid values. ignore subcycling.      npart = self('npd')
                    '_nosubf' --> same as '_fnosub'; provided for convenience.

                accounting for subcycling means dividing by the subcycling factor,
                because less effort is spent on subcycled distributions.
        '''
        clock, time, settings = _match.groups()
        clock_per_timestep = self(f'{clock}_time') / self('nit_since_prev')
        # get npart
        if settings is None:
            settings = ''
        if 'f' in settings:
            npart = self('npd')
            if 'nosub' not in settings:
                npart = npart / self('subcycle')
        else:
            with self.using(fluid=None):
                npart = self('npd')
                if 'nosub' not in settings:
                    npart = npart / self('subcycle')
                npart = xarray_sum(npart, 'fluid')
        result = clock_per_timestep / npart
        if time == 'dt':
            result = self.record_units(result / self.input_deck['dt'])
        elif time == 'timestep':
            pass  # already divided by timestep.
        else:
            raise NotImplementedError(f'coding error, expected time="timestep" or "dt", got {time!r}')
        return result


    # # # CONVERTING BETWEEN TIMERS.DAT <--> SECONDS # # #

    @known_var(deps=['run_time', 'steps_seconds'])
    def get_timer2seconds(self):
        '''conversion factor from timers.dat units to seconds.
        E.g. timer2seconds = 0.01 <--> 1 timer unit = 0.01 seconds <--> 1 second = 100 timer units.

        timer2seconds = steps_seconds / sum(run_time)
        Might vary from machine to machine. Seems to be 0.01 on Frontera (plus small rounding errors).
        '''
        return self('steps_seconds') / self('run_time').sum()

    @known_var(deps=['timer2seconds'])
    def get_seconds2timer(self):
        '''conversion factor from seconds to timers.dat units.
        E.g. seconds2timer = 100 <--> 1 timer unit = 0.01 seconds <--> 1 second = 100 timer units.

        seconds2timer = sum(run_time) / steps_seconds
        Might vary from machine to machine. Seems to be 100 on Frontera (plus small rounding errors).
        '''
        return 1 / self('timer2seconds')


    # # # TRYING TO "GUESS" RUNTIME PER TIMESTEP # # #

    # stats from some (~100) 2D runs I (sam) have on my machine,
    #   when computing 'log10_meant_cpu_seconds_per_each'.
    #   Excluded runs with ncells_sim <= 128**2 (because small runs timing isn't as important to get right)
    #       and runs with nx >= 20 (because I don't have enough runs to learn dependence with nsubdomains < max,
    #           and max nsubdomains is when nx = 8. Closest-to-optimal safe nsubdomains is when nx = 16.)
    CPU_SECONDS_PER_EACH_INFO = {
        'timers': ['vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'],
        'timer_shorthand': ['vadv', 'xadv', 'charge', 'collect', 'efield', 'output'],
        'stats': {
            'min': [-7.60274492, -8.08801379, -8.29066597, -6.4194993 , -6.45974422, -9.13447354],
            'mean': [-7.14767658, -7.71036947, -8.16571132, -5.41090113, -5.84219558, -8.26272736],
            'median': [-7.14165344, -7.7622515 , -8.18222028, -5.4868952 , -5.83410388, -8.29974968],
            'max': [-6.64987625, -6.85562141, -7.85642911, -4.26655371, -5.16467001, -6.96257236],
            'std': [0.1618933 , 0.24986124, 0.07723055, 0.39052264, 0.27228036, 0.41827969],
        },
        # arr.polyfit(log10_n_nodes, degree=1) where arr==log10_meant_cpu_seconds_per_each.
        'polyfit_n_nodes': {
            'coeff': {   # coeffs for each degree. Direct result of polyfit.
                1: [ 0.07367062,  0.33485588,  0.07118243,  0.98491967,  0.76134439,  0.42652909],
                0: [-7.23738158, -8.11810654, -8.25238656, -6.61018799, -6.76924614, -8.78209026],
            },
            'std': {   # stddev on coeffs for each degree. Larger stddev --> less certainty on coeff result.
                1: [0.05228419, 0.07328888, 0.02404283, 0.07154953, 0.03530289, 0.12873761],
                0: [0.06593647, 0.09242584, 0.03032081, 0.09023231, 0.04452106, 0.16235315],
            },
        },
        # which timers we should actually use polyfit_n_nodes for, when guessing runtime.
        #   default is ['collect', 'efield'] because those have smaller std & (max-min) after detrending
        #   (and since the fit is in logspace, that should directly cause smaller error in predictions).
        'polyfit_n_nodes_apply': ['collect', 'efield'],
    }

    @known_pattern(r'guess_cpu_seconds_(all|vadv|xadv|charge|collect|efield|output)_per_each(_safe|)',
                   deps=['n_nodes'])
    def get_guess_cpu_seconds_per_each(self, var, *, _match=None):
        '''guess log10(cpu seconds per (particle, per) grid cell, per timestep), for the indicated timer.
        if 'safe' included in var, uses mean + stddev instead of just mean, to get a conservative estimate.
        Result excludes initialization time (before iterations begin).

        timer options:
            'all': result will be a Dataset of results from all timers, with timer names as data_vars.
                    (timer names are 'vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'.)
            'vadv', 'xadv', 'charge', 'collect', 'efield', 'output': result for corresponding timer.

        if timer is 'all', return sum of the results from other timers.

        These results are per particle per grid cell per timestep:
            'vadv time', 'xadv time', 'charge', 'output'.
        These results are per particle per timestep:
            'collect', 'efield'.

        The guess is informed by self.CPU_SECONDS_PER_EACH_INFO,
            which contains numerical values derived from some simulations run on Frontera.
        Assumes 'output' is not very expensive.
            The "output" timer was not tested across a large variety of 'nout'.
            (trying to detrend with respect to nout didn't affect stddev significantly.)
        '''
        here, safe = _match.groups()
        INFO = self.CPU_SECONDS_PER_EACH_INFO
        if here == 'all':
            result = {}
            for i, timer_shorthand in enumerate(INFO['timer_shorthand']):
                timer = INFO['timers'][i]
                result[timer] = self(f'guess_cpu_seconds_{timer_shorthand}_per_each{safe}')
            return xr.Dataset(result)
        # else, getting result for an individual timer.
        i = INFO['timer_shorthand'].index(here)
        timer = INFO['timers'][i]
        if timer in INFO['polyfit_n_nodes_apply']:
            log_n_nodes = self('log10_n_nodes')
            pfit = INFO['polyfit_n_nodes']
            coeff1 = pfit['coeff'][1][i]
            coeff0 = pfit['coeff'][0][i]
            if safe:
                coeff1 = coeff1 + pfit['std'][1][i]
                coeff0 = coeff0 + pfit['std'][0][i]
            guess = coeff1 * log_n_nodes + coeff0
        else:
            guess = INFO['stats']['mean'][i]
            if safe:
                guess = guess + INFO['stats']['std'][i]
        # above, got log10(guess). Exponentiate and return:
        return xr.DataArray(10**guess)

    @known_pattern(r'guess_cpu_seconds_(all|sum|vadv|xadv|charge|collect|efield|output)_per_(timestep|ct)(_safe|)',
                   deps=[{1: 'cpu_per_each2{group1}'}, {2: 'guess_cpu_seconds_all_per_each{group2}'}])
    def get_guess_cpu_seconds_per_timestep_or_ct(self, var, *, _match=None):
        '''guess cpu seconds required per (cell, per) timestep, for the indicated timer.
        if 'safe' included in var, uses mean + stddev instead of just mean, to get a conservative estimate.
        Use 'ct' to indicate 'per cell per timestep'.
            (result using 'timestep') == ncells_sim * (result using 'ct').
        Result excludes initialization time (before iterations begin).

        timer options:
            'all': result will be a Dataset of results from all timers, with timer names as data_vars.
                    (timer names are 'vadv time', 'xadv time', 'charge', 'collect', 'efield', 'output'.)
            'sum': result will be the sum of the results from other timers.
            'vadv', 'xadv', 'charge', 'collect', 'efield', 'output': result for corresponding timer.

        Assumes 'output' is not very expensive; it was not tested as much as the others.
        The guess is informed by self.CPU_SECONDS_PER_EACH_INFO.
        For more details see: guess_cpu_seconds_per_each.
        '''
        here, ct, safe = _match.groups()
        INFO = self.CPU_SECONDS_PER_EACH_INFO
        if here == 'all' or here == 'sum':
            result = {}
            for i, timer_shorthand in enumerate(INFO['timer_shorthand']):
                timer = INFO['timers'][i]
                result[timer] = self(f'guess_cpu_seconds_{timer_shorthand}_per_{ct}{safe}')
            if here == 'all':
                return xr.Dataset(result)
            else:  # here == 'sum'
                return sum(result.values())
        # else, getting result for an individual timer.
        i = INFO['timer_shorthand'].index(here)
        timer = INFO['timers'][i]
        guess_each = self(f'guess_cpu_seconds_{here}_per_each{safe}')
        each2timestep = self(f'cpu_per_each2{ct}')[timer]
        return guess_each * each2timestep

    @known_pattern(r'guess_cpu_seconds(|_unsafe)',
                   deps=['guess_cpu_seconds_sum_per_timestep_safe'])  # [TODO] deps based on group0
    def get_guess_cpu_seconds(self, var, *, _match=None):
        '''guess run cost, in cpu seconds. nt * guess_cpu_seconds_sum_per_timestep_safe
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).
        '''
        unsafe, = _match.groups()
        safe = '' if unsafe else '_safe'
        return self.input_deck['nt'] * self(f'guess_cpu_seconds_sum_per_timestep{safe}')

    @known_pattern(r'guess_runtime_seconds(|_unsafe)',
                   deps=['n_processors', {0: 'guess_cpu_seconds{group0}'}])
    def get_guess_runtime_seconds(self, var, *, _match=None):
        '''guess run cost, in wall clock time [seconds]. guess_cpu_seconds / n_processors.
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).

        Compare directly with self('steps_seconds').
        '''
        unsafe, = _match.groups()
        return self(f'guess_cpu_seconds{unsafe}') / self('n_processors')

    @known_pattern(r'guess_node_hours(|_unsafe)',
                   deps=['tasks_per_node', {0: 'guess_cpu_seconds{group0}'}])
    def get_guess_node_hours(self, var, *, _match=None):
        '''guess run cost, in node hours. guess_cpu_seconds * seconds2hours / tasks_per_node
        (seconds2hours = 1/3600)
        Uses 'safe' guess (unless '_unsafe' included in name) but excludes initialization time.
            Might want to add a safety factor (e.g. assume 20% larger than this time).
        '''
        unsafe, = _match.groups()
        seconds2hours = 1 / 3600
        return self(f'guess_cpu_seconds{unsafe}') * seconds2hours / self('tasks_per_node')



"""
Package Purpose: managing the various formulae for quantities for PlasmaCalculator.
"""

from .collisions import CollisionsLoader, CrossTable
from .patterns import (
    # loaders
    ParenthesisLoader,
    BasicArithmeticLoader, BasicDerivativeLoader,
    CachesLoader,
    ComparisonLoader,
    BlurLoader, FFTLoader,
    FluidsLoader,
    PolyfitLoader,
    StatsLoader,
    VectorArithmeticLoader, VectorDerivativeLoader,
    # other things
    magnitude, unit_vector, angle_xy, angle_xy_to_hat,
    dot_product, cross_product, cross_component,
    take_perp_to, take_parallel_to,
)
from .bases import (
    AllBasesLoader, SimpleDerivedLoader, DirectBasesLoader,
    BASE_QUANTS, SIMPLE_DERIVED_QUANTS,
)
from .bperp import BperpLoader
from .caching import CachedQuantity, VarCache
from .constants import ConstantsLoader
from .direct_loader import DirectLoader
from . import help as _help_module
from .masks import MaskLoader
from .plasma_drifts import PlasmaDriftsLoader
from .plasma_heating import PlasmaHeatingLoader
from .plasma_parameters import PlasmaParametersLoader
from .plasma_stats import PlasmaStatsLoader
from .quasineutral import QuasineutralLoader
from .timescales import TimescalesLoader

# parent class & other tools, for the other loaders.
# Probably won't use these directly, outside of this subpackage.
from .quantity_loader import (
    MetaQuantTracking, QuantityLoader,
)
from .quantity_tools import (
    LoadableQuantity, LoadableVar, LoadablePattern,
    MatchedQuantity, MatchedVar, MatchedPattern,
    Pattern,
    DecoratingCalcs, DecoratingPatterns, DecoratingSetters,
    CallDepthMixin,
)



"""
File Purpose: tools related to scatter plots (possibly animated)
"""

import matplotlib.lines as mpl_lines
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .labels import xarray_title_plot_node
from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings
from .plot_tools import (
    set_margins_if_provided, set_lims_if_provided,
    infer_movie_dim, infer_xy_dims,
) 
from ..defaults import DEFAULTS
from ..errors import (
    InputError, InputConflictError,
    PlottingAmbiguityError,
)
from ..tools import (
    alias, alias_key_of, alias_child, simple_property,
    UNSET,
    pcAccessor,
    xarray_fill_coords, xarray_ensure_dims,
    xarray_nondim_coords, xarray_dims_coords,
)


''' --------------------- ScatterPlotElement & XarrayScatter --------------------- '''

@PlotSettings.format_docstring()
class ScatterPlotElement(MoviePlotElement):
    '''scatter plot.

    array: xarray.DataArray, probably ndim=1.
        the data to be plotted.
        if ndim != 1, must provide `x`; all other dims will be stacked.
    x: None or str
        the coordinate to use for x-axis values.
        None --> if array ndim==1, use the (first 1D) coord associated with the 1 dim,
            (or just use the dim itself, if no such associated coord).
            if array ndim!=1, crash with PlottingAmbiguityError.

    aliases, provide at most one of each pair:
        s, markersize: `s` from plt.scatter. size of markers.
        facecolors, facecolor: `facecolors` from plt.scatter. Color of marker faces in plot.
        edgecolors, edgecolor: `edgecolors` from plt.scatter. Color of marker edges in plot.
        linewidths, linewidth: `linewidths` from plt.scatter. Width of marker edges in plot.

    label: {label}
    {kw_ax_margin}
    '''
    def __init__(self, array, x=None, *, ax=None,
                 s=UNSET, markersize=UNSET,
                 facecolors=UNSET, facecolor=UNSET,
                 edgecolors=UNSET, edgecolor=UNSET,
                 linewidths=UNSET, linewidth=UNSET,
                 **kw_super):
        self._update_kw_from_aliases(kw_super, 's', s=s, markersize=markersize)
        self._update_kw_from_aliases(kw_super, 'facecolors', facecolors=facecolors, facecolor=facecolor)
        self._update_kw_from_aliases(kw_super, 'edgecolors', edgecolors=edgecolors, edgecolor=edgecolor)
        self._update_kw_from_aliases(kw_super, 'linewidths', linewidths=linewidths, linewidth=linewidth)
        super().__init__(**kw_super)
        self._ax_init = ax
        self._x_init = x  # <-- might help with debugging at some point
        self.x = self._infer_x(x, array)
        self.array = self._ensure_x_dim_in_array(array)
        self.init_scatter()

    array = alias_key_of('data', 'array', doc='''array for plot. Internally, stored at self.data['array']''')
    scatter = simple_property('_scatter', doc='''scatter plot object (matplotlib.collections.PathCollection)''')
    ax = alias_child('scatter', 'axes', if_no_child=None, doc='''the axes containing this scatter plot''')
    fig = alias_child('ax', 'figure', if_no_child=None, doc='''the figure containing this scatter plot''')

    def _update_kw_from_aliases(self, kw, key, **aliases):
        '''update kw[key] = value implied from aliases.
        if multiple aliases' values are not UNSET, raise InputConflictError.
        if all aliases' values are UNSET, do not update kw[key].
        returns the updated value, or UNSET of no update was made.
        '''
        exists = [k for k, v in aliases.items() if v is not UNSET]
        if len(exists) == 0:
            return UNSET
        elif len(exists) == 1:
            val = aliases[exists[0]]
            kw[key] = val
            return val
        else:  # len(exists) >= 2
            raise InputConflictError(f'Cannot provide multiple values for {key!r}: {exists}')

    def _infer_x(self, x, array):
        '''set self.x and self.array. See help(type(self)) for more details.'''
        if x is None:
            if array.ndim == 1:
                dim_x = array.dims[0]
                dims_coords = xarray_dims_coords(array, include_dims_as_coords=False)
                for x_opt in dims_coords[dim_x]:
                    if array[x_opt].ndim == 1:
                        x = x_opt
                        break
                else:  # didn't break, no decent coord option found
                    x = dim_x
            else:
                raise PlottingAmbiguityError(f"array.ndim={array.ndim} but x=None. Cannot infer x.")
        else:
            if x not in array.coords:
                raise InputError(f"x={x!r} not in array.coords={set(array.coords)}")
        return x

    def _ensure_x_dim_in_array(self, array):
        '''Return array if self.x already a dim, else array.expand_dims(self.x)'''
        if self.x in array.dims:
            return array
        else:
            return array.expand_dims(self.x)

    # # # PLOTTING THIS ELEMENT # # #
    def init_scatter(self):
        '''initialize the scatter plot; actually plot the data (probably).
        stores plotted object in self.scatter and returns self.scatter.

        if the array is all nan, plot nothing and return None.
        otherwise, set self._plotted_at_least_once = True.
        '''
        array = self.array
        # plot settings / kwargs / bookkeeping
        kw_plot = self.plot_settings.get_mpl_kwargs('xarray.DataArray.plot.scatter')
        if 'label' in kw_plot:
            kw_plot['label'] = kw_plot['label'].format(**xarray_nondim_coords(array))
        # make plot
        if array.isnull().all():
            self._plotted_at_least_once = False
            return None
        else:
            self.scatter = array.plot.scatter(ax=self._ax_init, x=self.x, **kw_plot)
            self._plotted_at_least_once = True
            # formatting
            set_margins_if_provided(self.plot_settings, ax=self.ax)
            set_lims_if_provided(self.plot_settings, ax=self.ax)
            return self.scatter

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''update the plot using data['array'].
        return the list of all updated matplotlib Artist objects;
            Probably [self.scatter], but might be empty list (if array is all nan).
        (if self.init_scatter() didn't make a plot due to all nan initial values, use that instead.)
        '''
        # [TODO][EFF] if same number of points, update existing points instead of making new ones.
        array = data['array']
        array = self._ensure_x_dim_in_array(array)
        if not getattr(self, '_plotted_at_least_once', False):
            # if never plotted anything yet due to all nans, try init_scatter() instead
            self.array = array
            result = self.init_scatter()
            return [] if result is None else [result]
        # remove existing scatter (if any). Remember some old values first.
        ax_old, kw_old = self._remember_then_remove_old()
        # plot new scatter
        if array.isnull().all():
            return []
        else:
            kw_plot = self.plot_settings.get_mpl_kwargs('xarray.DataArray.plot.scatter')
            kw_plot.update(add_labels=False)  # no adjusting existing labels here.
            kw_plot.update(kw_old)
            self.scatter = array.plot.scatter(ax=ax_old, x=self.x, **kw_plot)
            return [self.scatter]

    def _remember_then_remove_old(self):
        '''remember any relevant details from the old scatter plot (before removing it).
        saves self._ax_old and self._kw_old, then does old.remove(), where old=self.scatter.
        If self.scatter doesn't exist, return the saved _ax_old and _kw_old.

        returns (self._ax_old, self._kw_old)
        '''
        old = getattr(self, 'scatter', None)
        if old is None:
            assert hasattr(self, '_ax_old'), "ScatterPlotElement.update_data called before init_scatter?"
            assert hasattr(self, '_kw_old'), "ScatterPlotElement.update_data called before init_scatter?"
        else:
            self._ax_old = self.ax
            kw_old = {}
            if 'c' not in self.plot_settings:
                # if all points same color, keep it; don't use ax color_cycler.
                old_colors = old.get_facecolors()
                if len(old_colors)>0 and np.all(old_colors == old_colors[0]):
                    kw_old.update(color=old_colors[0])
            old.remove()
            del self.scatter
            self._kw_old = kw_old
        return (self._ax_old, self._kw_old)

    # # # CONENIENCE # # #
    def legend_handle(self, *, label=UNSET):
        '''returns a matplotlib.lines.Line2D suitable for use as a handle in a legend.
        Caution: not yet compatible with scatter plots with varying marker colors / sizes.
        
        label: UNSET or str
            if not UNSET, use this as the label, instead of self.plot_settings['label']

        Example:
            handle1 = self.legend_handle()
            handle2 = other_scatter_plot.legend_handle()
            plt.legend(handles=[handle1, handle2])

        One reason to consider using this is that self might plot 0 points on an axis sometimes,
            so if creating the legend "automatically" instead, then it wouldn't show up.
        '''
        kw = self.plot_settings.get_mpl_kwargs('mpl.lines.Line2D', label=label)
        # some aliases (because matplotlib is annoyingly inconsistent...):
        kw_scatter = self.plot_settings.get_mpl_kwargs('xarray.DataArray.plot.scatter')
        if 'linewidths' in kw_scatter:
            kw['markeredgewidth'] = kw_scatter['linewidths']
        if 's' in kw_scatter:
            # Line2D markersize actually means sqrt(s)...
            kw['markersize'] = np.sqrt(kw_scatter['s'])
        # color-related aliases (need to check if marker is filled...):
        marker = kw.get('marker', plt.rcParams['scatter.marker'])
        if marker in mpl_lines.Line2D.filled_markers:
            # facecolor means facecolor. (nice.)
            # edgecolor means edgecolor. (also nice.)
            if 'facecolors' in kw_scatter:
                kw['markerfacecolor'] = kw_scatter['facecolors']
            if 'edgecolors' in kw_scatter:
                kw['markeredgecolor'] = kw_scatter['edgecolors']
        else:
            # facecolor means edgecolor. (ew.)
            # edgecolor gets ignored entirely. (ewww.)
            if 'facecolors' in kw_scatter:
                kw['markeredgecolor'] = kw_scatter['facecolors']
            if 'edgecolors' in kw_scatter:
                pass  # don't do anything, matplotlib doesn't want to know about it.
        # always overwrite:
        kw['linestyle'] = ''  # no line
        return mpl_lines.Line2D([], [], **kw)


@pcAccessor.register('scatter')
@PlotSettings.format_docstring()
class XarrayScatter(MoviePlotNode):  # [TODO] refactor PlotDimsMixin to use it here.
    '''MoviePlotNode of scatter.
    stores a ScatterPlotElement & has methods for plotting & updating the plot.
    "scatter" refers to a matplotlib.collections.PathCollection object, e.g. the result of plt.scatter.

    array: xarray.DataArray, probably ndim=2.
        the data to be plotted. if ndim=1, can still plot, but nothing to animate.
        For ndim >= 3, consider manually iterating over the 3rd+ dim(s),
            putting multing scatter plots onto the same axes,
            making a movie=MovieOrganizerNode() and doing movie.add_child(s) for each plot.
            [TODO] add a class (or just a function?) to handle this sort of thing automatically.
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None -> infer from array & any other provided dimensions.
    x: None or str
        if provided, tells dimensions for the x plot axes.
        None -> infer from array & any other provided dimensions.
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.

    init_plot: {init_plot}
    label: {label}

    {kw_ax_margin}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    '''
    element_cls = ScatterPlotElement

    def __init__(self, array, t=None, *, x=None, ax=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 label=PlotSettings.default('label'),
                 **kw_super):
        if isinstance(array, xr.Dataset):
            array = xarray_as_array(array)
        array = xarray_fill_coords(array, need=[x, t])
        array = xarray_ensure_dims(array, ({x, t} - {None}), promote_dims_if_needed=True)
        self.array = array
        self._ax_init = ax
        # infer dims ([TODO]: refactor plot dims code & offload this logic to PlotDimsMixin.)
        if array.ndim != 1:
            t = infer_movie_dim(array.dims, t, fail_ok=True)
        if (array.ndim != 1) and (t is None):
            errmsg = 't=None but is required (when array.ndim != 2). But cannot infer t from data provided.'
            raise PlottingAmbiguityError(errmsg)
        if x is None:
            x = infer_xy_dims(array.dims, x=x, y=None, exclude=[t], fail_ok=True)[0]
        if x is None:
            errmsg = 'x=None but is required. But cannot infer x from data provided.'
            raise PlottingAmbiguityError(errmsg)
        self.t = t
        self.x = x
        kw_super.update(init_plot=init_plot, label=label)
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    t_plot_dim = alias('t')

    ax = alias_child('obj', 'ax', if_special_child={None: None, UNSET: None},
        doc='''mpl.axes.Axes where this XarrayScatter is plotted, or None if not plotted.''')

    fig = alias_child('obj', 'fig', if_special_child={None: None, UNSET: None},
        doc='''figure where this XarrayScatter is plotted, or None if not plotted.''')

    scatter = alias_child('obj', 'scatter', if_special_child={None: None, UNSET: None},
        doc='''mpl.collections.PathCollection object of this XarrayScatter, or None if not plotted.''')

    legend_handle = alias_child('obj', 'legend_handle',
        doc='''alias to self.obj.legend_handle;
        method which returns a Line2D object suitable for use as a handle in a legend.
        (implied handles are sometimes okay, but this is useful in case self is sometimes all-nan.)''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the ScatterPlotElement at self.obj'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs(f'pc.ScatterPlotElement')
        # -- determine ylims (based on the entire array, not just this frame)
        # DISABLED. Usually I scatter on top of existing plots, and don't want scatter to touch ylim.
        #   Workaround: use XarrayLine() instead. [TODO] optionally enable?
        # d = kw_plot
        # robust = d.get('robust', False)
        # ylim = d.pop('ylim', None)
        # ymin, ymax = (None, None) if (ylim is None) else ylim
        # ymin, ymax = calculate_vlims(self.array, vmin=ymin, vmax=ymax, robust=robust)
        # d['ylim'] = (ymin, ymax)
        # >> actually plot the scatter <<
        self.obj = self.element_cls(data['array'], ax=self._ax_init, x=self.x, **kw_plot)
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        # bookkeeping
        self.frame = frame

    def get_data_at_frame(self, frame):
        '''returns {'array': array at this frame}, properly transposed & ready for plotting.'''
        t = self.t_plot_dim
        result = self.array
        if t is not None:  # there is a dimension for the time axis.
            result = result.isel({t: frame})
        return {'array': result}

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t_plot_dim
        if t is None:
            return 0
        return len(self.array.coords[t])

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is None:
            plt.title('')  # remove xarray's default title.
        else:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.array, title, t=self.t_plot_dim,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging


''' --------------------- scatter-related methods --------------------- '''

@pcAccessor.register('scatter_max')
def xarray_scatter_max(array, x, y, *, t=None, **kw_scatter):
    '''use XarrayScatter to plt.scatter() a single marker, at the argmax of array.
    default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MAX}

    array: xarray.DataArray
        array whose max will be marked.
    x: str
        coordinate to use for x-axis values.
    y: str
        coordinate to use for y-axis values.
    t: None or str
        coordinate to use for "time" (iterate across movie frames).
        None --> probably can't animate; might show all points in 1 frame.

    [TODO] optionally infer x, y, t automatically if not provided directly.

    additional kwargs go to XarrayScatter
    returns XarrayScatter object.
    '''
    xdim = x
    ydim = y
    argmax = array.argmax((xdim, ydim))
    x = array[xdim].isel({xdim: argmax[xdim]})
    y = array[ydim].isel({ydim: argmax[ydim]})
    kw = {**DEFAULTS.PLOT.SCATTER_STYLE,
          **DEFAULTS.PLOT.SCATTER_MAX,
          **kw_scatter}
    scatarr = y.assign_coords({xdim: x})
    return XarrayScatter(scatarr, x=xdim, t=t, **kw)

@pcAccessor.register('scatter_min')
def xarray_scatter_min(array, x, y, *, t=None, **kw_scatter):
    '''use XarrayScatter to plt.scatter() a single marker, at the argmin of array.
    default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MIN}

    array: xarray.DataArray
        array whose min will be marked.
    x: str
        coordinate to use for x-axis values.
    y: str
        coordinate to use for y-axis values.
    t: None or str
        coordinate to use for "time" (iterate across movie frames).
        None --> probably can't animate; might show all points in 1 frame.

    [TODO] optionally infer x, y, t automatically if not provided directly.

    additional kwargs go to XarrayScatter
    returns XarrayScatter object.
    '''
    xdim = x
    ydim = y
    argmin = array.argmin((xdim, ydim))
    x = array[xdim].isel({xdim: argmin[xdim]})
    y = array[ydim].isel({ydim: argmin[ydim]})
    kw = {**DEFAULTS.PLOT.SCATTER_STYLE,
          **DEFAULTS.PLOT.SCATTER_MIN,
          **kw_scatter}
    scatarr = y.assign_coords({xdim: x})
    return XarrayScatter(scatarr, x=xdim, t=t, **kw)

@pcAccessor.register('scatter_where')
def xarray_scatter_where(arr, condition=None, *, x, y, t=None, **kw_scatter):
    '''use XarrayScatter to plt.scatter() markers wherever arr==True.
    Roughly equivalent: arr[y].where(arr).pc.scatter(x=x, t=t, **kw_scatter)

    arr: xarray.DataArray
        array of bools, or array to be provided to condition if condition is callable.
    condition: None, array of bools, or callable
        None --> use condition = `arr`, and `arr` must be array of bools.
        array --> ignore arr entirely. `condition` must be array of bools.
        callable --> use condition = condition(`arr`), which must return array of bools.
    x: str
        coordinate to use for x-axis values.
    y: str
        coordinate to use for y-axis values.
    t: None or str
        coordinate to use for "time" (iterate across movie frames).
        None --> probably can't animate; might show all points in 1 frame.

    [TODO] optionally infer x, y, t automatically if not provided directly.

    additional kwargs go to XarrayScatter
    returns XarrayScatter object.
    '''
    if condition is None:
        where = arr
    elif callable(condition):
        where = condition(arr)
    else:
        where = condition
    where = xarray_ensure_dims(where, ({x, y, t} - {None}), promote_dims_if_needed=True)
    scatarr = where[y].where(where)
    return XarrayScatter(scatarr, x=x, t=t, **kw_scatter)



"""
File Purpose: loader for tfbi-related quantities
"""
import os

from .tfbi_solver import TfbiSolver, _paramdocs_tfbi_solving
from ..addon_tools import register_addon_loader_if
from ...defaults import DEFAULTS
from ...errors import FluidValueError, InputMissingError, MemorySizeError
from ...quantities import QuantityLoader
from ...tools import (
    format_docstring,
    simple_property,
    xarray_min, xarray_grid,
    xarray_load, xarray_save,
    xarray_min_coord_where,
)


''' --------------------- TfbiLoader --------------------- '''

@register_addon_loader_if(DEFAULTS.ADDONS.LOAD_TFBI)
@format_docstring(**_paramdocs_tfbi_solving)
class TfbiLoader(QuantityLoader):
    '''quantities related to the Thermal Farley Buneman Instability.
    
    NOTE: for simple calculations, consider using maindims_means=True!

    To solve TFBI theory, you can use the following pattern:
        {solving_pattern}
        
        Notes:
            {solving_pattern_notes}
    '''
    tfbi_solver_cls = TfbiSolver

    def tfbi_mask(self, *, kappae=1, ionfrac=1e-3, kappai=1, set=True):
        '''set & return self.mask appropriate for TFBI.
        kappae: None or number, default 1
            lower limit for kappae; mask points with kappae smaller than this value.
            (kappae = |qe| |B| / (me nuen))
            Internally, loaded as 'kappa' with fluid='e'.
            TFBI probably only matters when electrons are magnetized --> kappae >> 1.
            Applying TFBI theory to kappae masked points would still be fine,
                but will probably always say "instability doesn't grow there".
            None --> no mask on kappae
        ionfrac: None or number
            upper limit for ionfrac; mask points with ionfrac larger than this value.
            (ionfrac = ne / ntotal)
            Internally, loaded as 'SF_ionfrac' if available,
                else ne/(ne+n_neutral), with ne from self('n', fluid=self.fluids.get_electron()).
                (Note: checked in Bifrost: SF_ionfrac uses SF_n = sum of element densities.
                    SF_n does not include ne.
                    The formula SF_ionfrac = ne/(ne+nn) uses ne as a proxy for sum(ni),
                        which will be okay unless there are twice+ ionized species.)
            TFBI assumes weakly ionized.
            E_un0 also assumes weakly ionized, when self.assume_un='u'.
            Applying TFBI theory to ionfrac masked points would be a big issue,
                as it could lead to many false positives,
                where physical effects not included in the theory damp out the TFBI.
            None --> no mask on ionfrac.
        kappai: None or number, default 1
            upper limit for min kappai; mask points with all kappai larger than this value.
            (kappai = |qi| |B| / (mi nuin))
            Internally, loaded as 'kappa' with fluid=ions, then take min across fluids.
            TFBI probably only matters when at least 1 ion species is demagnetized --> kappai < 1.
            Applying TFBI theory to kappai masked points would still be fine,
                but will probably always say "instability doesn't grow there".
            None --> no mask on kappai.
        set: bool
            whether to set self.mask = result.
            if False, only returns the result, without also setting self.mask.
        '''
        if all(x is None for x in (kappae, ionfrac, kappai)):
            raise InputMissingError('At least one must be non-None: kappae, ionfrac, kappai.')
        with self.using(masking=False):
            if ionfrac is not None:
                if self.has_var('SF_ionfrac'):
                    ionfrac_vals = self('SF_ionfrac')
                else:
                    ne = self('n', fluid=self.fluids.get_electron())
                    nn = self('n_neutral')
                    ionfrac_vals = ne / (ne + nn)
            if kappae is not None:
                kappae_vals = self('kappa', fluid=self.fluids.get_electron())
            if kappai is not None:
                kappai_vals = self('kappa', fluid=self.fluids.ions())
                kappai_vals = xarray_min(kappai_vals, dim='fluid')
        mask = None
        if ionfrac is not None:
            mask_ionfrac = (ionfrac_vals > ionfrac)
            mask = mask_ionfrac if mask is None else (mask | mask_ionfrac)
        if kappae is not None:
            mask_kappae = (kappae_vals < kappae)
            mask = mask_kappae if mask is None else (mask | mask_kappae)
        if kappai is not None:
            mask_kappai = (kappai_vals > kappai)
            mask = mask_kappai if mask is None else (mask | mask_kappai)
        if set:
            self.mask = mask
            mask = self.mask  # <-- may have been altered slightly due to mask attr.
        return mask

    def tfbi_ds(self, ions=None, *, all=True, output_mask=True, **kw_get_var):
        '''returns Dataset of all the values needed & relevant to TFBI theory.
        Equivalent: self('tfbi_all', fluid=[electron, *ions], masking=True, ...)
        
        ions: None or specifier of multiple fluids (e.g. slice, or list of strs)
            list of ions to use. None --> self.fluids.ions()
        all: bool
            whether to include 'tfbi_all', or only 'tfbi_inputs'.
            With only 'tfbi_inputs', the theory is still solvable, but harder to inspect later.
        output_mask: bool
            whether to store_mask in results, if self.masking (and self.mask is not None)
        additional kwargs passed to self(...)
        '''
        if ions is None:
            ions = self.fluids.ions()
        fluid = [self.fluids.get_electron(), *ions]
        tfbi_var = 'tfbi_all' if all else 'tfbi_inputs'
        return self(tfbi_var, fluid=fluid, output_mask=output_mask, **kw_get_var)

    @format_docstring(tfbi_solver_docs=TfbiSolver.__doc__, sub_ntab=1)
    def tfbi_solver(self, ions=None, **kw_solver):
        '''return TfbiSolver object for solving TFBI theory based on values in self.
        all inputs here get passed to TfbiSolver. Equivalent: TfbiSolver(self, ...).

        docs for TfbiSolver copied below for convenience:
        -------------------------------------------------
        {tfbi_solver_docs}
        '''
        return self.tfbi_solver_cls(self, ions=ions, **kw_solver)

    # # # LOADING TFBI INPUTS & RELATED VARS # # #
    TFBI_VARS = ['mod_B', 'E_un0_perpmod_B', 'kB', 'T_n', 'm_n',  # "global" scalars
                    'm', 'nusn', 'skappa', 'eqperp_ldebye']  # scalars which depend on fluid.

    # extra vars, relevant to TFBI theory, but not necessary.
    TFBI_EXTRAS = ['SF_n', 'eps0', 'abs_qe', # "global" scalars
                    'n', 'n_n', 'eqperp_lmfp',
                    # 'tfbi_fscale_rel',  # this one turned out to be irrelevant --> exclude by default
                    ]   # scalars which depend on fluid.

    @known_var(deps=TFBI_VARS)
    def get_tfbi_inputs(self, **kw_get_vars):
        '''returns xarray.Dataset of values to input to the tfbi theory.
        "global" scalars (no dependence on component nor fluid)
            'mod_B': |magnetic field|
            'E_un0_perpmag_B': |E_un0 perp to B|. E_un0 = electric field in u_neutral=0 frame.
            'kB': boltzmann constant. kB * T = temperature in energy units.
            'T_n': temperature of neutrals.
            'm_n': mass of neutrals.
        scalars which depend on fluid. Note: checks self.fluid, not self.fluids.
            'm': mass of all non-neutral fluids
            'nusn': collision frequency between fluid and neutrals.
            'skappa': signed magnetization parameter; q |B| / (m nusn)
            'eqperp_ldebye': each fluid's debye length at its "equilibrium" temperature,
                        after considering zeroth order heating due to E_un0_perpmag_B.

        Results depend on self.fluid. May want to call as self('tfbi_inputs', fluid=CHARGED).
        '''
        if any(f.is_neutral() for f in self.iter_fluid()):
            errmsg = ('get_tfbi_inputs expects self.fluid to be charged fluids only,\n'
                      f'but it includes neutrals: {[f for f in self.iter_fluid() if f.is_neutral()]}')
            raise FluidValueError(errmsg)
        # [TODO][EFF] improve efficiency by avoiding redundant calculations,
        #    e.g. B is calculated separately for mod_B, E_un0_perpmag_B, and skappa,
        #    while E_un0 is calculated separately for E_un0_perpmag_B and eqperp_ldebye.
        tfbi_vars = self.TFBI_VARS
        return self(tfbi_vars, **kw_get_vars)

    @known_var(deps=TFBI_EXTRAS)
    def get_tfbi_extras(self, **kw_get_vars):
        '''returns xarray.Dataset of values relevant to TFBI theory but not necessary for inputs.
        Currently this just includes:
            'eqperp_lmfp': each fluid's collisional mean free path at its "equilibrium" temperature,
                        after considering zeroth order heating due to E_un0_perpmag_B.
            'SF_n': sum of number densities of all species (including neutrals)
            'n': number densities of each specie in self.fluid.
            'n_n': number density of neutral fluid.
            'n*kappa': number density times kappa.
                    TFBI dispersion relation terms scale with n*kappa for each fluid,
                    so this quantity roughly estimates the relative importance of each fluid.

        Results depend on self.fluid. May want to call as self('tfbi_extras', fluid=CHARGED).
        '''
        if any(f.is_neutral() for f in self.iter_fluid()):
            errmsg = ('get_tfbi_extras expects self.fluid to be charged fluids only,\n'
                      f'but it includes neutrals: {[f for f in self.iter_fluid() if f.is_neutral()]}')
            raise FluidValueError(errmsg)
        tfbi_extras = self.TFBI_EXTRAS
        kw_get_vars.setdefault('missing_vars', 'ignore')  # it's okay if some extras are un-gettable.
        return self(tfbi_extras, **kw_get_vars)

    @known_var(deps=['tfbi_inputs', 'tfbi_extras'])
    def get_tfbi_all(self, **kw_get_vars):
        '''returns xarray.Dataset of values relevant to TFBI theory.
        This includes tfbi_inputs (required for theory) and tfbi_extras (optional)

        Results depend on self.fluid. May want to call as self('tfbi_all', fluid=CHARGED).
        '''
        return self(['tfbi_inputs', 'tfbi_extras'], **kw_get_vars)

    @known_var(deps=['n*kappa'])
    def get_tfbi_fscale(self):
        '''tfbi_fscale = n * kappa
        tfbi dispersion relation sums terms proportional to n * kappa, for each fluid.
        '''
        return self('n*kappa')

    @known_var(deps=['tfbi_fscale'])
    def get_tfbi_fscale_rel(self):
        '''tfbi_fscale_rel = tfbi_fscale(this fluid) / tfbi_fscale(electrons).'''
        return self('tfbi_fscale') / self('tfbi_fscale', fluid=self.fluids.get_electron())

    # # # LOADING TFBI SOLUTION # # #
    @known_var(deps=['tfbi_inputs'])
    def get_tfbi_omega(self, *, kw_tfbi_solve=dict(), **kw_tfbi_solver):
        '''Thermal Farley Buneman Instability roots with largest imaginary part at each point in self.
        Equivalent: self.tfbi_solver(**kw_tfbi_solver).solve(**kw_solve)['omega'].

        Can provide kwargs, e.g. self('tfbi_omega', ions=['H_II', 'C_II'], kw_tfbi_solve=dict(ncpu=1)).

        For more control, use self.tfbi_solver() directly.
        For even more control, use the pattern described in help(self.tfbi_solver_cls).

        Recommended: consider using 'tfbi_omega_ds' instead of 'tfbi_omega'.
            'tfbi_omega_ds' gives the full Dataset of all values relevant to the solution.
            'tfbi_omega' just gives the DataArray of omega, which is harder to inspect later.
        '''
        kw_tfbi_solver.setdefault('tfbi_all', False)  # dropping ds0 later anyways so don't load it.
        solver = self.tfbi_solver(**kw_tfbi_solver)
        kw_tfbi_solve = kw_tfbi_solve.copy()
        kw_tfbi_solve.setdefault('return_ds', False)  # just return omega, not the full ds.
        # (if user specified return_ds=True, then they will get the full ds, but that's okay.)
        return solver(**kw_tfbi_solve)

    @known_var(deps=['tfbi_all'])
    def get_tfbi_omega_ds(self, *, kw_tfbi_solve=dict(), **kw_tfbi_solver):
        '''Thermal Farley Buneman Instability solution at each point in self.
        Equivalent: self.tfbi_solver(**kw_tfbi_solver).solve(**kw_solve).

        Can provide kwargs, e.g. self('tfbi_omega_ds', ions=['H_II', 'C_II'], kw_tfbi_solve=dict(ncpu=1)).

        For more control, use self.tfbi_solver() directly.
        For even more control, use the pattern described in help(self.tfbi_solver_cls).
        '''
        kw_tfbi_solver.setdefault('tfbi_all', True)
        solver = self.tfbi_solver(**kw_tfbi_solver)
        return solver(**kw_tfbi_solve)


    # # # --- SETTING VALUES; KNOWN SETTERS --- # # #
    # used when using set_var.

    @known_setter(aliases=['mag_B'])
    def set_mod_B(self, value, **kw):
        '''set mod_B to this value. Also sets mag_B, mod2_B, and mag2_B.'''
        # [TODO] pattern handling for setters, for mod; see issue #5 on git for more info.
        #   (implementing pattern handling should make this function obsolete / mostly obsolete.)
        self.set_var_internal('mod_B', value, ['snap'], **kw, ukey='b_field')
        self.set_var_internal('mag_B', value, ['snap'], **kw, ukey='b_field')
        value2 = value**2
        self.set_var_internal('mod2_B', value2, ['snap'], **kw, ukey='b_field2')
        self.set_var_internal('mag2_B', value2, ['snap'], **kw, ukey='b_field2')

    @known_setter(aliases=['E_un0_perpmag_B'])
    def set_E_un0_perpmod_B(self, value, **kw):
        '''set E_un0_perpmod_B to this value. Also sets E_un0_perpmag_B.'''
        self.set_var_internal('E_un0_perpmag_B', value, ['snap'], **kw, ukey='e_field')
        self.set_var_internal('E_un0_perpmod_B', value, ['snap'], **kw, ukey='e_field')


    # # # --- COMPUTING E_THRESH --- # # #

    _default_grid_size = (
            DEFAULTS.ADDONS.TFBI_EBSPEED_LOGMIN,
            DEFAULTS.ADDONS.TFBI_EBSPEED_LOGMAX,
            DEFAULTS.ADDONS.TFBI_EBSPEED_LOGSTEP
    )
    cls_behavior_attrs.register('tfbi_EBspeed_grid_size', default=_default_grid_size)
    tfbi_EBspeed_grid_size = simple_property('_tfbi_EBspeed_grid_size', default=_default_grid_size,
        doc=f'''(logmin, logmax, logstep) for self('tfbi_EBspeed_grid'); log is base 10.
        Default: DEFAULTS.ADDONS.TFBI_EBSPEED_LOGMIN, LOGMAX, LOGSTEP; (default: {_default_grid_size}).''')
    del _default_grid_size

    @known_var
    def get_tfbi_EBspeed_grid(self):
        '''return a 1D grid of EBspeed values with constant logstep.
        determines logmin, logmax, logstep (base 10) from self.tfbi_EBspeed_grid_size.
        result's name & EBspeed grid dim is always 'EBspeed'.
        '''
        logmin, logmax, logstep = self.tfbi_EBspeed_grid_size
        return xarray_grid(10**logmin, 10**logmax, logstep=logstep, name='EBspeed')

    def _tfbi_vs_EBspeed_file(self):
        '''returns expected abspath to file for storing tfbi solution across EBspeed grid.
        Depends on current value of self.tfbi_EBspeed_grid_size.
        Does not check whether file exists already
            (might use this to find existing file, or to determine filepath for saving result.)
        Result is like:
            {self.unique_notes_dirname}/_pc_tfbi/EBspeed_{logmin:.4g}_{logmax:.4g}_{logstep:.4g}.pcxarr
        '''
        logmin, logmax, logstep = self.tfbi_EBspeed_grid_size
        basename = f'EBspeed_{logmin:.4g}_{logmax:.4g}_{logstep:.4g}.pcxarr'
        return os.path.join(self.unique_notes_dirname, '_pc_tfbi', basename)

    @format_docstring(_mbytes_max=DEFAULTS.ADDONS.TFBI_EBSPEED_INPUTS_MBYTES_MAX)
    def solve_tfbi_vs_EBspeed(self, *, Mbytes_max=True, cache='caches', **kw_solve):
        '''solve tfbi across EBspeed grid from self('tfbi_EBspeed_grid').
        CAUTION: the implementation here might self.set('E_un0_perpmod_B', self('mod_B') * EBspeed)
        [TODO] avoid changing self.setvars... (or, at least, restore previous self.setvars afterwards.)
        Suggestion: use self.setvars.clear() after calling this function.

        Mbytes_max: bool or number
            maxmimum allowed data size [in MB] of self('tfbi_inputs'), before setting EBspeed grid.
            (helps prevent accidental requests to solve too many points at once.)
            (ignored if `cache` implies to load from existing cached file.)
            True --> use default: DEFAULTS.ADDONS.TFBI_EBSPEED_INPUTS_MBYTES_MAX (default: {_mbytes_max}).
            False or None --> no maximum
        cache: 'caches', 'cache', 'cached', or False
            controls how & whether to handle cache the result.
            cached results go to self._tfbi_vs_EBspeed_file(); default:
                {{self.unique_notes_dirname}}/_pc_tfbi/EBspeed_{{logmin:.4g}}_{{logmax:.4g}}_{{logstep:.4g}}.pcxarr
            where logmin, logmax, logstep are from self.tfbi_EBspeed_grid_size
            'caches' --> read from saved file if it exists. Else, solve and save to file.
            'cache' --> solve and save to file. Saved file must not exist yet (else, crash).
            'cached' --> read from saved file. Saved file must exist (else, crash).
            False --> solve and return result, but do not save to file nor check if file exists.
        additional kwargs are passed to self.tfbi_solver().solve(**kw)
        '''
        # check for existing file
        if cache:
            filename = self._tfbi_vs_EBspeed_file()
            if cache == 'caches':
                if os.path.exists(filename):
                    return xarray_load(filename)
                # else, cache it later
            elif cache == 'cached':
                if os.path.exists(filename):
                    return xarray_load(filename)
                else:
                    raise FileNotFoundError(f'cache="cached" but _tfbi_vs_EBspeed_file() does not exist: {filename!r}.')
            elif cache == 'cache':
                if os.path.exists(filename):
                    raise FileExistsError(f'cache="cache" but _tfbi_vs_EBspeed_file() already exists: {filename!r}.')
                # else, cache it later
        # check size of tfbi_inputs
        if Mbytes_max is not False and Mbytes_max is not None:
            if Mbytes_max is True:
                Mbytes_max = DEFAULTS.ADDONS.TFBI_EBSPEED_INPUTS_MBYTES_MAX
            try:
                tfbi_inputs = self('tfbi_inputs', array_MBmax=Mbytes_max)
            except MemorySizeError:
                inputs_Mbytes = f'an individual tfbi_input array, alone, is already larger than {Mbytes_max}'
                too_big = True
            else:
                inputs_Mbytes = f'{tfbi_inputs.pc.nMbytes:.4g}'
                too_big = tfbi_inputs.pc.nMbytes > Mbytes_max
            if too_big:
                errmsg = (f'self("tfbi_inputs").pc.nMbytes ({inputs_Mbytes} MB) > Mbytes_max ({Mbytes_max}).'
                          '\nConsider setting larger Mbytes_max (or DEFAULTS.ADDONS.TFBI_EBSPEED_INPUTS_MBYTES_MAX),'
                          f' or reducing the number of points being considered here.')
                raise MemorySizeError(errmsg)
        # set E based on EBspeed grid
        EBspeed = self('tfbi_EBspeed_grid')
        self.set('E_un0_perpmag_B', EBspeed * self('mod_B'))
        # solve TFBI
        solver = self.tfbi_solver()
        result = solver.solve(**kw_solve)
        # cache
        if cache:
            xarray_save(result, filename)
        return result

    @known_var
    def get_tfbi_vs_EBspeed(self):
        '''return tfbi solution across EBspeed grid. Load saved result if it exists, else save result to file.
        CAUTION: the implementation here might self.set('E_un0_perpmod_B', self('mod_B') * EBspeed)
        [TODO] avoid changing self.setvars... (or, at least, restore previous self.setvars afterwards.)

        CAUTION: the implementation here assumes self.tfbi_EBspeed_grid_size is enough to uniquely specify the result;
            e.g. if there is a different result at each snapshot of self, that will not be understood here.

        Equivalent: self.solve_tfbi_vs_EBspeed(cache='caches')
        '''
        verbose = getattr(self, 'verbose', True)
        return self.solve_tfbi_vs_EBspeed(cache='caches', verbose=verbose)

    @known_var
    def get_cached_tfbi_vs_EBspeed(self):
        '''return a previously-computed tfbi solution across EBspeed grid.
        Expects solution to live in _tfbi_vs_EBspeed_file(); default:
            {self.unique_notes_dirname}/_pc_tfbi/EBspeed_{logmin:.4g}_{logmax:.4g}_{logstep:.4g}.pcxarr

        CAUTION: the implementation here assumes self.tfbi_EBspeed_grid_size is enough to uniquely specify the result;
            e.g. if there is a different result at each snapshot of self, that will not be understood here.
        '''
        filename = self._tfbi_vs_EBspeed_file()
        if not os.path.exists(filename):
            errmsg = (f'tfbi_vs_EBspeed solution not found: {filename!r}'
                      f'\nConsider running self.solve_tfbi_vs_EBspeed() to compute it, '
                      'after downsampling somehow (e.g. via self.maindims_means=True, self.snap=0).'
                      '\nCAUTION: self.solve_tfbi_vs_EBspeed() might self.set("E_un0_perpmod_B"); '
                      f'consider using self.copy() first, or making a new {type(self).__name__} object after.')
            raise FileNotFoundError(errmsg)
        return xarray_load(filename)

    cls_behavior_attrs.register('tfbi_growth_thresh', default=DEFAULTS.ADDONS.TFBI_GROWTH_THRESH)
    tfbi_growth_thresh = simple_property('_tfbi_growth_thresh', default=DEFAULTS.ADDONS.TFBI_GROWTH_THRESH,
        doc=f'''threshold for confirming "yes there is tfbi growth predicted here", during self('tfbi_E_thresh').
        growth must be larger than (not equal to) this value, to confirm growth.
        0.0 is the theoretical threshold. A small positive value (e.g. 0.001) helps to avoid tiny errors.
        The default is DEFAULTS.ADDONS.TFBI_GROWTH_THRESH (default: {DEFAULTS.ADDONS.TFBI_GROWTH_THRESH}).''')

    @known_var(deps=['cached_tfbi_vs_EBspeed'])
    def get_tfbi_EBspeed_thresh(self):
        '''threshold EBspeed for TFBI to grow. NaN if no growth predicted across the EBspeed grid considered.
        
        Assumes user has already run self.solve_tfbi_vs_EBspeed().
        If not, consider doing:
            copied = self.copy()
            copied.set_attrs(maindims_means=True, snap=0)  # or some other way to downsample...
            copied.solve_tfbi_vs_EBspeed()

        equivalent to tfbi_vs_EBspeed.pc.min_coord_where('EBspeed', tfbi_vs_EBspeed.it.growth_kmax()>0)
        '''
        tfbi_vs_EBspeed = self('cached_tfbi_vs_EBspeed')
        growth_kmax = tfbi_vs_EBspeed.it.growth_kmax()
        return xarray_min_coord_where(tfbi_vs_EBspeed, 'EBspeed', growth_kmax > 0)

    @known_var(deps=['tfbi_EBspeed_thresh', 'mod_B'])
    def get_tfbi_E_thresh(self):
        '''threshold E_un0_perpmag_B for TFBI to grow. NaN if no growth predicted across the EBspeed grid considered.
        Equivalent: self('tfbi_EBspeed_thresh') * self('mod_B').

        Assumes user has already run self.solve_tfbi_vs_EBspeed().
        See self.help('tfbi_EBspeed_thresh') for more details.
        '''
        EBspeed = self('tfbi_EBspeed_thresh')
        return EBspeed * self('mod_B')
